import sys
from _typeshed import SupportsRead
from typing import IO, Any, NamedTuple, overload
from typing_extensions import final

if sys.platform != "win32":
    _chtype = str | bytes | int

    # ACS codes are only initialized after initscr is called
    ACS_BBSS: int
    ACS_BLOCK: int
    ACS_BOARD: int
    ACS_BSBS: int
    ACS_BSSB: int
    ACS_BSSS: int
    ACS_BTEE: int
    ACS_BULLET: int
    ACS_CKBOARD: int
    ACS_DARROW: int
    ACS_DEGREE: int
    ACS_DIAMOND: int
    ACS_GEQUAL: int
    ACS_HLINE: int
    ACS_LANTERN: int
    ACS_LARROW: int
    ACS_LEQUAL: int
    ACS_LLCORNER: int
    ACS_LRCORNER: int
    ACS_LTEE: int
    ACS_NEQUAL: int
    ACS_PI: int
    ACS_PLMINUS: int
    ACS_PLUS: int
    ACS_RARROW: int
    ACS_RTEE: int
    ACS_S1: int
    ACS_S3: int
    ACS_S7: int
    ACS_S9: int
    ACS_SBBS: int
    ACS_SBSB: int
    ACS_SBSS: int
    ACS_SSBB: int
    ACS_SSBS: int
    ACS_SSSB: int
    ACS_SSSS: int
    ACS_STERLING: int
    ACS_TTEE: int
    ACS_UARROW: int
    ACS_ULCORNER: int
    ACS_URCORNER: int
    ACS_VLINE: int
    ALL_MOUSE_EVENTS: int
    A_ALTCHARSET: int
    A_ATTRIBUTES: int
    A_BLINK: int
    A_BOLD: int
    A_CHARTEXT: int
    A_COLOR: int
    A_DIM: int
    A_HORIZONTAL: int
    A_INVIS: int
    if sys.version_info >= (3, 7):
        A_ITALIC: int
    A_LEFT: int
    A_LOW: int
    A_NORMAL: int
    A_PROTECT: int
    A_REVERSE: int
    A_RIGHT: int
    A_STANDOUT: int
    A_TOP: int
    A_UNDERLINE: int
    A_VERTICAL: int
    BUTTON1_CLICKED: int
    BUTTON1_DOUBLE_CLICKED: int
    BUTTON1_PRESSED: int
    BUTTON1_RELEASED: int
    BUTTON1_TRIPLE_CLICKED: int
    BUTTON2_CLICKED: int
    BUTTON2_DOUBLE_CLICKED: int
    BUTTON2_PRESSED: int
    BUTTON2_RELEASED: int
    BUTTON2_TRIPLE_CLICKED: int
    BUTTON3_CLICKED: int
    BUTTON3_DOUBLE_CLICKED: int
    BUTTON3_PRESSED: int
    BUTTON3_RELEASED: int
    BUTTON3_TRIPLE_CLICKED: int
    BUTTON4_CLICKED: int
    BUTTON4_DOUBLE_CLICKED: int
    BUTTON4_PRESSED: int
    BUTTON4_RELEASED: int
    BUTTON4_TRIPLE_CLICKED: int
    # Darwin ncurses doesn't provide BUTTON5_* constants
    if sys.version_info >= (3, 10) and sys.platform != "darwin":
        BUTTON5_PRESSED: int
        BUTTON5_RELEASED: int
        BUTTON5_CLICKED: int
        BUTTON5_DOUBLE_CLICKED: int
        BUTTON5_TRIPLE_CLICKED: int
    BUTTON_ALT: int
    BUTTON_CTRL: int
    BUTTON_SHIFT: int
    COLOR_BLACK: int
    COLOR_BLUE: int
    COLOR_CYAN: int
    COLOR_GREEN: int
    COLOR_MAGENTA: int
    COLOR_RED: int
    COLOR_WHITE: int
    COLOR_YELLOW: int
    ERR: int
    KEY_A1: int
    KEY_A3: int
    KEY_B2: int
    KEY_BACKSPACE: int
    KEY_BEG: int
    KEY_BREAK: int
    KEY_BTAB: int
    KEY_C1: int
    KEY_C3: int
    KEY_CANCEL: int
    KEY_CATAB: int
    KEY_CLEAR: int
    KEY_CLOSE: int
    KEY_COMMAND: int
    KEY_COPY: int
    KEY_CREATE: int
    KEY_CTAB: int
    KEY_DC: int
    KEY_DL: int
    KEY_DOWN: int
    KEY_EIC: int
    KEY_END: int
    KEY_ENTER: int
    KEY_EOL: int
    KEY_EOS: int
    KEY_EXIT: int
    KEY_F0: int
    KEY_F1: int
    KEY_F10: int
    KEY_F11: int
    KEY_F12: int
    KEY_F13: int
    KEY_F14: int
    KEY_F15: int
    KEY_F16: int
    KEY_F17: int
    KEY_F18: int
    KEY_F19: int
    KEY_F2: int
    KEY_F20: int
    KEY_F21: int
    KEY_F22: int
    KEY_F23: int
    KEY_F24: int
    KEY_F25: int
    KEY_F26: int
    KEY_F27: int
    KEY_F28: int
    KEY_F29: int
    KEY_F3: int
    KEY_F30: int
    KEY_F31: int
    KEY_F32: int
    KEY_F33: int
    KEY_F34: int
    KEY_F35: int
    KEY_F36: int
    KEY_F37: int
    KEY_F38: int
    KEY_F39: int
    KEY_F4: int
    KEY_F40: int
    KEY_F41: int
    KEY_F42: int
    KEY_F43: int
    KEY_F44: int
    KEY_F45: int
    KEY_F46: int
    KEY_F47: int
    KEY_F48: int
    KEY_F49: int
    KEY_F5: int
    KEY_F50: int
    KEY_F51: int
    KEY_F52: int
    KEY_F53: int
    KEY_F54: int
    KEY_F55: int
    KEY_F56: int
    KEY_F57: int
    KEY_F58: int
    KEY_F59: int
    KEY_F6: int
    KEY_F60: int
    KEY_F61: int
    KEY_F62: int
    KEY_F63: int
    KEY_F7: int
    KEY_F8: int
    KEY_F9: int
    KEY_FIND: int
    KEY_HELP: int
    KEY_HOME: int
    KEY_IC: int
    KEY_IL: int
    KEY_LEFT: int
    KEY_LL: int
    KEY_MARK: int
    KEY_MAX: int
    KEY_MESSAGE: int
    KEY_MIN: int
    KEY_MOUSE: int
    KEY_MOVE: int
    KEY_NEXT: int
    KEY_NPAGE: int
    KEY_OPEN: int
    KEY_OPTIONS: int
    KEY_PPAGE: int
    KEY_PREVIOUS: int
    KEY_PRINT: int
    KEY_REDO: int
    KEY_REFERENCE: int
    KEY_REFRESH: int
    KEY_REPLACE: int
    KEY_RESET: int
    KEY_RESIZE: int
    KEY_RESTART: int
    KEY_RESUME: int
    KEY_RIGHT: int
    KEY_SAVE: int
    KEY_SBEG: int
    KEY_SCANCEL: int
    KEY_SCOMMAND: int
    KEY_SCOPY: int
    KEY_SCREATE: int
    KEY_SDC: int
    KEY_SDL: int
    KEY_SELECT: int
    KEY_SEND: int
    KEY_SEOL: int
    KEY_SEXIT: int
    KEY_SF: int
    KEY_SFIND: int
    KEY_SHELP: int
    KEY_SHOME: int
    KEY_SIC: int
    KEY_SLEFT: int
    KEY_SMESSAGE: int
    KEY_SMOVE: int
    KEY_SNEXT: int
    KEY_SOPTIONS: int
    KEY_SPREVIOUS: int
    KEY_SPRINT: int
    KEY_SR: int
    KEY_SREDO: int
    KEY_SREPLACE: int
    KEY_SRESET: int
    KEY_SRIGHT: int
    KEY_SRSUME: int
    KEY_SSAVE: int
    KEY_SSUSPEND: int
    KEY_STAB: int
    KEY_SUNDO: int
    KEY_SUSPEND: int
    KEY_UNDO: int
    KEY_UP: int
    OK: int
    REPORT_MOUSE_POSITION: int
    _C_API: Any
    version: bytes
    def baudrate() -> int: ...
    def beep() -> None: ...
    def can_change_color() -> bool: ...
    def cbreak(__flag: bool = ...) -> None: ...
    def color_content(__color_number: int) -> tuple[int, int, int]: ...
    # Changed in Python 3.8.8 and 3.9.2
    if sys.version_info >= (3, 8):
        def color_pair(pair_number: int) -> int: ...
    else:
        def color_pair(__color_number: int) -> int: ...

    def curs_set(__visibility: int) -> int: ...
    def def_prog_mode() -> None: ...
    def def_shell_mode() -> None: ...
    def delay_output(__ms: int) -> None: ...
    def doupdate() -> None: ...
    def echo(__flag: bool = ...) -> None: ...
    def endwin() -> None: ...
    def erasechar() -> bytes: ...
    def filter() -> None: ...
    def flash() -> None: ...
    def flushinp() -> None: ...
    if sys.version_info >= (3, 9):
        def get_escdelay() -> int: ...
        def get_tabsize() -> int: ...

    def getmouse() -> tuple[int, int, int, int, int]: ...
    def getsyx() -> tuple[int, int]: ...
    def getwin(__file: SupportsRead[bytes]) -> _CursesWindow: ...
    def halfdelay(__tenths: int) -> None: ...
    def has_colors() -> bool: ...
    if sys.version_info >= (3, 10):
        def has_extended_color_support() -> bool: ...

    def has_ic() -> bool: ...
    def has_il() -> bool: ...
    def has_key(__key: int) -> bool: ...
    def init_color(__color_number: int, __r: int, __g: int, __b: int) -> None: ...
    def init_pair(__pair_number: int, __fg: int, __bg: int) -> None: ...
    def initscr() -> _CursesWindow: ...
    def intrflush(__flag: bool) -> None: ...
    def is_term_resized(__nlines: int, __ncols: int) -> bool: ...
    def isendwin() -> bool: ...
    def keyname(__key: int) -> bytes: ...
    def killchar() -> bytes: ...
    def longname() -> bytes: ...
    def meta(__yes: bool) -> None: ...
    def mouseinterval(__interval: int) -> None: ...
    def mousemask(__newmask: int) -> tuple[int, int]: ...
    def napms(__ms: int) -> int: ...
    def newpad(__nlines: int, __ncols: int) -> _CursesWindow: ...
    def newwin(__nlines: int, __ncols: int, __begin_y: int = ..., __begin_x: int = ...) -> _CursesWindow: ...
    def nl(__flag: bool = ...) -> None: ...
    def nocbreak() -> None: ...
    def noecho() -> None: ...
    def nonl() -> None: ...
    def noqiflush() -> None: ...
    def noraw() -> None: ...
    def pair_content(__pair_number: int) -> tuple[int, int]: ...
    def pair_number(__attr: int) -> int: ...
    def putp(__string: bytes) -> None: ...
    def qiflush(__flag: bool = ...) -> None: ...
    def raw(__flag: bool = ...) -> None: ...
    def reset_prog_mode() -> None: ...
    def reset_shell_mode() -> None: ...
    def resetty() -> None: ...
    def resize_term(__nlines: int, __ncols: int) -> None: ...
    def resizeterm(__nlines: int, __ncols: int) -> None: ...
    def savetty() -> None: ...
    if sys.version_info >= (3, 9):
        def set_escdelay(__ms: int) -> None: ...
        def set_tabsize(__size: int) -> None: ...

    def setsyx(__y: int, __x: int) -> None: ...
    def setupterm(term: str | None = ..., fd: int = ...) -> None: ...
    def start_color() -> None: ...
    def termattrs() -> int: ...
    def termname() -> bytes: ...
    def tigetflag(__capname: str) -> int: ...
    def tigetnum(__capname: str) -> int: ...
    def tigetstr(__capname: str) -> bytes | None: ...
    def tparm(
        __str: bytes,
        __i1: int = ...,
        __i2: int = ...,
        __i3: int = ...,
        __i4: int = ...,
        __i5: int = ...,
        __i6: int = ...,
        __i7: int = ...,
        __i8: int = ...,
        __i9: int = ...,
    ) -> bytes: ...
    def typeahead(__fd: int) -> None: ...
    def unctrl(__ch: _chtype) -> bytes: ...
    def unget_wch(__ch: int | str) -> None: ...
    def ungetch(__ch: _chtype) -> None: ...
    def ungetmouse(__id: int, __x: int, __y: int, __z: int, __bstate: int) -> None: ...
    def update_lines_cols() -> None: ...
    def use_default_colors() -> None: ...
    def use_env(__flag: bool) -> None: ...

    class error(Exception): ...

    @final
    class _CursesWindow:
        encoding: str
        @overload
        def addch(self, ch: _chtype, attr: int = ...) -> None: ...
        @overload
        def addch(self, y: int, x: int, ch: _chtype, attr: int = ...) -> None: ...
        @overload
        def addnstr(self, str: str, n: int, attr: int = ...) -> None: ...
        @overload
        def addnstr(self, y: int, x: int, str: str, n: int, attr: int = ...) -> None: ...
        @overload
        def addstr(self, str: str, attr: int = ...) -> None: ...
        @overload
        def addstr(self, y: int, x: int, str: str, attr: int = ...) -> None: ...
        def attroff(self, __attr: int) -> None: ...
        def attron(self, __attr: int) -> None: ...
        def attrset(self, __attr: int) -> None: ...
        def bkgd(self, __ch: _chtype, __attr: int = ...) -> None: ...
        def bkgdset(self, __ch: _chtype, __attr: int = ...) -> None: ...
        def border(
            self,
            ls: _chtype = ...,
            rs: _chtype = ...,
            ts: _chtype = ...,
            bs: _chtype = ...,
            tl: _chtype = ...,
            tr: _chtype = ...,
            bl: _chtype = ...,
            br: _chtype = ...,
        ) -> None: ...
        @overload
        def box(self) -> None: ...
        @overload
        def box(self, vertch: _chtype = ..., horch: _chtype = ...) -> None: ...
        @overload
        def chgat(self, attr: int) -> None: ...
        @overload
        def chgat(self, num: int, attr: int) -> None: ...
        @overload
        def chgat(self, y: int, x: int, attr: int) -> None: ...
        @overload
        def chgat(self, y: int, x: int, num: int, attr: int) -> None: ...
        def clear(self) -> None: ...
        def clearok(self, yes: int) -> None: ...
        def clrtobot(self) -> None: ...
        def clrtoeol(self) -> None: ...
        def cursyncup(self) -> None: ...
        @overload
        def delch(self) -> None: ...
        @overload
        def delch(self, y: int, x: int) -> None: ...
        def deleteln(self) -> None: ...
        @overload
        def derwin(self, begin_y: int, begin_x: int) -> _CursesWindow: ...
        @overload
        def derwin(self, nlines: int, ncols: int, begin_y: int, begin_x: int) -> _CursesWindow: ...
        def echochar(self, __ch: _chtype, __attr: int = ...) -> None: ...
        def enclose(self, __y: int, __x: int) -> bool: ...
        def erase(self) -> None: ...
        def getbegyx(self) -> tuple[int, int]: ...
        def getbkgd(self) -> tuple[int, int]: ...
        @overload
        def getch(self) -> int: ...
        @overload
        def getch(self, y: int, x: int) -> int: ...
        @overload
        def get_wch(self) -> int | str: ...
        @overload
        def get_wch(self, y: int, x: int) -> int | str: ...
        @overload
        def getkey(self) -> str: ...
        @overload
        def getkey(self, y: int, x: int) -> str: ...
        def getmaxyx(self) -> tuple[int, int]: ...
        def getparyx(self) -> tuple[int, int]: ...
        @overload
        def getstr(self) -> bytes: ...
        @overload
        def getstr(self, n: int) -> bytes: ...
        @overload
        def getstr(self, y: int, x: int) -> bytes: ...
        @overload
        def getstr(self, y: int, x: int, n: int) -> bytes: ...
        def getyx(self) -> tuple[int, int]: ...
        @overload
        def hline(self, ch: _chtype, n: int) -> None: ...
        @overload
        def hline(self, y: int, x: int, ch: _chtype, n: int) -> None: ...
        def idcok(self, flag: bool) -> None: ...
        def idlok(self, yes: bool) -> None: ...
        def immedok(self, flag: bool) -> None: ...
        @overload
        def inch(self) -> int: ...
        @overload
        def inch(self, y: int, x: int) -> int: ...
        @overload
        def insch(self, ch: _chtype, attr: int = ...) -> None: ...
        @overload
        def insch(self, y: int, x: int, ch: _chtype, attr: int = ...) -> None: ...
        def insdelln(self, nlines: int) -> None: ...
        def insertln(self) -> None: ...
        @overload
        def insnstr(self, str: str, n: int, attr: int = ...) -> None: ...
        @overload
        def insnstr(self, y: int, x: int, str: str, n: int, attr: int = ...) -> None: ...
        @overload
        def insstr(self, str: str, attr: int = ...) -> None: ...
        @overload
        def insstr(self, y: int, x: int, str: str, attr: int = ...) -> None: ...
        @overload
        def instr(self, n: int = ...) -> bytes: ...
        @overload
        def instr(self, y: int, x: int, n: int = ...) -> bytes: ...
        def is_linetouched(self, __line: int) -> bool: ...
        def is_wintouched(self) -> bool: ...
        def keypad(self, yes: bool) -> None: ...
        def leaveok(self, yes: bool) -> None: ...
        def move(self, new_y: int, new_x: int) -> None: ...
        def mvderwin(self, y: int, x: int) -> None: ...
        def mvwin(self, new_y: int, new_x: int) -> None: ...
        def nodelay(self, yes: bool) -> None: ...
        def notimeout(self, yes: bool) -> None: ...
        @overload
        def noutrefresh(self) -> None: ...
        @overload
        def noutrefresh(self, pminrow: int, pmincol: int, sminrow: int, smincol: int, smaxrow: int, smaxcol: int) -> None: ...
        @overload
        def overlay(self, destwin: _CursesWindow) -> None: ...
        @overload
        def overlay(
            self, destwin: _CursesWindow, sminrow: int, smincol: int, dminrow: int, dmincol: int, dmaxrow: int, dmaxcol: int
        ) -> None: ...
        @overload
        def overwrite(self, destwin: _CursesWindow) -> None: ...
        @overload
        def overwrite(
            self, destwin: _CursesWindow, sminrow: int, smincol: int, dminrow: int, dmincol: int, dmaxrow: int, dmaxcol: int
        ) -> None: ...
        def putwin(self, __file: IO[Any]) -> None: ...
        def redrawln(self, __beg: int, __num: int) -> None: ...
        def redrawwin(self) -> None: ...
        @overload
        def refresh(self) -> None: ...
        @overload
        def refresh(self, pminrow: int, pmincol: int, sminrow: int, smincol: int, smaxrow: int, smaxcol: int) -> None: ...
        def resize(self, nlines: int, ncols: int) -> None: ...
        def scroll(self, lines: int = ...) -> None: ...
        def scrollok(self, flag: bool) -> None: ...
        def setscrreg(self, __top: int, __bottom: int) -> None: ...
        def standend(self) -> None: ...
        def standout(self) -> None: ...
        @overload
        def subpad(self, begin_y: int, begin_x: int) -> _CursesWindow: ...
        @overload
        def subpad(self, nlines: int, ncols: int, begin_y: int, begin_x: int) -> _CursesWindow: ...
        @overload
        def subwin(self, begin_y: int, begin_x: int) -> _CursesWindow: ...
        @overload
        def subwin(self, nlines: int, ncols: int, begin_y: int, begin_x: int) -> _CursesWindow: ...
        def syncdown(self) -> None: ...
        def syncok(self, flag: bool) -> None: ...
        def syncup(self) -> None: ...
        def timeout(self, delay: int) -> None: ...
        def touchline(self, start: int, count: int, changed: bool = ...) -> None: ...
        def touchwin(self) -> None: ...
        def untouchwin(self) -> None: ...
        @overload
        def vline(self, ch: _chtype, n: int) -> None: ...
        @overload
        def vline(self, y: int, x: int, ch: _chtype, n: int) -> None: ...
    if sys.version_info >= (3, 8):
        class _ncurses_version(NamedTuple):
            major: int
            minor: int
            patch: int
        ncurses_version: _ncurses_version
        window = _CursesWindow  # undocumented



import types
from _typeshed import StrPath
from typing import Any, Callable, Mapping, Sequence, TypeVar
from typing_extensions import ParamSpec

_T = TypeVar("_T")
_P = ParamSpec("_P")
_localtrace = Callable[[types.FrameType, str, Any], Callable[..., Any]]
_fileModuleFunction = tuple[str, str | None, str]

class CoverageResults:
    def __init__(
        self,
        counts: dict[tuple[str, int], int] | None = ...,
        calledfuncs: dict[_fileModuleFunction, int] | None = ...,
        infile: StrPath | None = ...,
        callers: dict[tuple[_fileModuleFunction, _fileModuleFunction], int] | None = ...,
        outfile: StrPath | None = ...,
    ) -> None: ...  # undocumented
    def update(self, other: CoverageResults) -> None: ...
    def write_results(self, show_missing: bool = ..., summary: bool = ..., coverdir: StrPath | None = ...) -> None: ...
    def write_results_file(
        self, path: StrPath, lines: Sequence[str], lnotab: Any, lines_hit: Mapping[int, int], encoding: str | None = ...
    ) -> tuple[int, int]: ...

class Trace:
    def __init__(
        self,
        count: int = ...,
        trace: int = ...,
        countfuncs: int = ...,
        countcallers: int = ...,
        ignoremods: Sequence[str] = ...,
        ignoredirs: Sequence[str] = ...,
        infile: StrPath | None = ...,
        outfile: StrPath | None = ...,
        timing: bool = ...,
    ) -> None: ...
    def run(self, cmd: str | types.CodeType) -> None: ...
    def runctx(
        self, cmd: str | types.CodeType, globals: Mapping[str, Any] | None = ..., locals: Mapping[str, Any] | None = ...
    ) -> None: ...
    def runfunc(self, func: Callable[_P, _T], *args: _P.args, **kw: _P.kwargs) -> _T: ...
    def file_module_function_of(self, frame: types.FrameType) -> _fileModuleFunction: ...
    def globaltrace_trackcallers(self, frame: types.FrameType, why: str, arg: Any) -> None: ...
    def globaltrace_countfuncs(self, frame: types.FrameType, why: str, arg: Any) -> None: ...
    def globaltrace_lt(self, frame: types.FrameType, why: str, arg: Any) -> None: ...
    def localtrace_trace_and_count(self, frame: types.FrameType, why: str, arg: Any) -> _localtrace: ...
    def localtrace_trace(self, frame: types.FrameType, why: str, arg: Any) -> _localtrace: ...
    def localtrace_count(self, frame: types.FrameType, why: str, arg: Any) -> _localtrace: ...
    def results(self) -> CoverageResults: ...



from typing import Any

RESULT_SUCCESS: int
RESULT_OPERATIONS_ERROR: int
RESULT_PROTOCOL_ERROR: int
RESULT_TIME_LIMIT_EXCEEDED: int
RESULT_SIZE_LIMIT_EXCEEDED: int
RESULT_COMPARE_FALSE: int
RESULT_COMPARE_TRUE: int
RESULT_AUTH_METHOD_NOT_SUPPORTED: int
RESULT_STRONGER_AUTH_REQUIRED: int
RESULT_RESERVED: int
RESULT_REFERRAL: int
RESULT_ADMIN_LIMIT_EXCEEDED: int
RESULT_UNAVAILABLE_CRITICAL_EXTENSION: int
RESULT_CONFIDENTIALITY_REQUIRED: int
RESULT_SASL_BIND_IN_PROGRESS: int
RESULT_NO_SUCH_ATTRIBUTE: int
RESULT_UNDEFINED_ATTRIBUTE_TYPE: int
RESULT_INAPPROPRIATE_MATCHING: int
RESULT_CONSTRAINT_VIOLATION: int
RESULT_ATTRIBUTE_OR_VALUE_EXISTS: int
RESULT_INVALID_ATTRIBUTE_SYNTAX: int
RESULT_NO_SUCH_OBJECT: int
RESULT_ALIAS_PROBLEM: int
RESULT_INVALID_DN_SYNTAX: int
RESULT_ALIAS_DEREFERENCING_PROBLEM: int
RESULT_INAPPROPRIATE_AUTHENTICATION: int
RESULT_INVALID_CREDENTIALS: int
RESULT_INSUFFICIENT_ACCESS_RIGHTS: int
RESULT_BUSY: int
RESULT_UNAVAILABLE: int
RESULT_UNWILLING_TO_PERFORM: int
RESULT_LOOP_DETECTED: int
RESULT_NAMING_VIOLATION: int
RESULT_OBJECT_CLASS_VIOLATION: int
RESULT_NOT_ALLOWED_ON_NON_LEAF: int
RESULT_NOT_ALLOWED_ON_RDN: int
RESULT_ENTRY_ALREADY_EXISTS: int
RESULT_OBJECT_CLASS_MODS_PROHIBITED: int
RESULT_AFFECT_MULTIPLE_DSAS: int
RESULT_OTHER: int
RESULT_LCUP_RESOURCES_EXHAUSTED: int
RESULT_LCUP_SECURITY_VIOLATION: int
RESULT_LCUP_INVALID_DATA: int
RESULT_LCUP_UNSUPPORTED_SCHEME: int
RESULT_LCUP_RELOAD_REQUIRED: int
RESULT_CANCELED: int
RESULT_NO_SUCH_OPERATION: int
RESULT_TOO_LATE: int
RESULT_CANNOT_CANCEL: int
RESULT_ASSERTION_FAILED: int
RESULT_AUTHORIZATION_DENIED: int
RESULT_E_SYNC_REFRESH_REQUIRED: int
RESULT_CODES: Any
DO_NOT_RAISE_EXCEPTIONS: Any



import _compression
import sys
import zlib
from _typeshed import ReadableBuffer, StrOrBytesPath
from io import FileIO
from typing import Any, Protocol, TextIO, overload
from typing_extensions import Literal

if sys.version_info >= (3, 8):
    __all__ = ["BadGzipFile", "GzipFile", "open", "compress", "decompress"]
else:
    __all__ = ["GzipFile", "open", "compress", "decompress"]

_ReadBinaryMode = Literal["r", "rb"]
_WriteBinaryMode = Literal["a", "ab", "w", "wb", "x", "xb"]
_OpenTextMode = Literal["rt", "at", "wt", "xt"]

READ: Literal[1]
WRITE: Literal[2]

class _ReadableFileobj(Protocol):
    def read(self, __n: int) -> bytes: ...
    def seek(self, __n: int) -> Any: ...
    # The following attributes and methods are optional:
    # name: str
    # mode: str
    # def fileno() -> int: ...

class _WritableFileobj(Protocol):
    def write(self, __b: bytes) -> Any: ...
    def flush(self) -> Any: ...
    # The following attributes and methods are optional:
    # name: str
    # mode: str
    # def fileno() -> int: ...

@overload
def open(
    filename: StrOrBytesPath | _ReadableFileobj,
    mode: _ReadBinaryMode = ...,
    compresslevel: int = ...,
    encoding: None = ...,
    errors: None = ...,
    newline: None = ...,
) -> GzipFile: ...
@overload
def open(
    filename: StrOrBytesPath | _WritableFileobj,
    mode: _WriteBinaryMode,
    compresslevel: int = ...,
    encoding: None = ...,
    errors: None = ...,
    newline: None = ...,
) -> GzipFile: ...
@overload
def open(
    filename: StrOrBytesPath,
    mode: _OpenTextMode,
    compresslevel: int = ...,
    encoding: str | None = ...,
    errors: str | None = ...,
    newline: str | None = ...,
) -> TextIO: ...
@overload
def open(
    filename: StrOrBytesPath | _ReadableFileobj | _WritableFileobj,
    mode: str,
    compresslevel: int = ...,
    encoding: str | None = ...,
    errors: str | None = ...,
    newline: str | None = ...,
) -> GzipFile | TextIO: ...

class _PaddedFile:
    file: _ReadableFileobj
    def __init__(self, f: _ReadableFileobj, prepend: bytes = ...) -> None: ...
    def read(self, size: int) -> bytes: ...
    def prepend(self, prepend: bytes = ...) -> None: ...
    def seek(self, off: int) -> int: ...
    def seekable(self) -> bool: ...

if sys.version_info >= (3, 8):
    class BadGzipFile(OSError): ...

class GzipFile(_compression.BaseStream):
    myfileobj: FileIO | None
    mode: Literal[1, 2]
    name: str
    compress: zlib._Compress
    fileobj: _ReadableFileobj | _WritableFileobj
    @overload
    def __init__(
        self,
        filename: StrOrBytesPath | None,
        mode: _ReadBinaryMode,
        compresslevel: int = ...,
        fileobj: _ReadableFileobj | None = ...,
        mtime: float | None = ...,
    ) -> None: ...
    @overload
    def __init__(
        self,
        *,
        mode: _ReadBinaryMode,
        compresslevel: int = ...,
        fileobj: _ReadableFileobj | None = ...,
        mtime: float | None = ...,
    ) -> None: ...
    @overload
    def __init__(
        self,
        filename: StrOrBytesPath | None,
        mode: _WriteBinaryMode,
        compresslevel: int = ...,
        fileobj: _WritableFileobj | None = ...,
        mtime: float | None = ...,
    ) -> None: ...
    @overload
    def __init__(
        self,
        *,
        mode: _WriteBinaryMode,
        compresslevel: int = ...,
        fileobj: _WritableFileobj | None = ...,
        mtime: float | None = ...,
    ) -> None: ...
    @overload
    def __init__(
        self,
        filename: StrOrBytesPath | None = ...,
        mode: str | None = ...,
        compresslevel: int = ...,
        fileobj: _ReadableFileobj | _WritableFileobj | None = ...,
        mtime: float | None = ...,
    ) -> None: ...
    @property
    def filename(self) -> str: ...
    @property
    def mtime(self) -> int | None: ...
    crc: int
    def write(self, data: ReadableBuffer) -> int: ...
    def read(self, size: int | None = ...) -> bytes: ...
    def read1(self, size: int = ...) -> bytes: ...
    def peek(self, n: int) -> bytes: ...
    @property
    def closed(self) -> bool: ...
    def close(self) -> None: ...
    def flush(self, zlib_mode: int = ...) -> None: ...
    def fileno(self) -> int: ...
    def rewind(self) -> None: ...
    def readable(self) -> bool: ...
    def writable(self) -> bool: ...
    def seekable(self) -> bool: ...
    def seek(self, offset: int, whence: int = ...) -> int: ...
    def readline(self, size: int | None = ...) -> bytes: ...

class _GzipReader(_compression.DecompressReader):
    def __init__(self, fp: _ReadableFileobj) -> None: ...
    def read(self, size: int = ...) -> bytes: ...

if sys.version_info >= (3, 8):
    def compress(data: bytes, compresslevel: int = ..., *, mtime: float | None = ...) -> bytes: ...

else:
    def compress(data: bytes, compresslevel: int = ...) -> bytes: ...

def decompress(data: bytes) -> bytes: ...



from .api import (
    CANCEL as CANCEL,
    NO_RETVAL as NO_RETVAL,
    contains as contains,
    listen as listen,
    listens_for as listens_for,
    remove as remove,
)
from .attr import RefCollection as RefCollection
from .base import Events as Events, dispatcher as dispatcher



"""Data structures and algorithms for boolean equations."""

import collections
import itertools

from pytype.pytd import pytd_utils

chain = itertools.chain.from_iterable


class BooleanTerm:
  """Base class for boolean terms."""

  __slots__ = ()

  def simplify(self, assignments):
    """Simplify this term, given a list of possible values for each variable.

    Args:
      assignments: A list of possible values for each variable. A dictionary
        mapping strings (variable name) to sets of strings (value names).

    Returns:
      A new BooleanTerm, potentially simplified.
    """
    raise NotImplementedError()

  def extract_pivots(self, assignments):
    """Find values for every variable that appears in this term.

    This finds all variables that appear in this term and limits them to the
    values they appear together with. For example, consider the equation
      t = v1 | (t = v2 & (t = v2 | t = v3))
    Here, t can be limited to [v1, v2]. (v3 is impossible.)

    Args:
      assignments: The current "upper bound", i.e. all values that are still
        possible for variables. Used for extracting pivots out of Eq(var, var).

    Returns:
      A dictionary mapping strings (variable names) to sets of strings (value
      or variable names).
    """
    raise NotImplementedError()

  def extract_equalities(self):
    """Find all equalities that appear in this term.

    Returns:
      A sequence of tuples of a string (variable name) and a string (value or
      variable name).
    """
    raise NotImplementedError()


class TrueValue(BooleanTerm):
  """Class for representing "TRUE"."""

  def simplify(self, assignments):
    return self

  def __repr__(self):
    return "TRUE"

  def __str__(self):
    return "TRUE"

  def extract_pivots(self, assignments):
    return {}

  def extract_equalities(self):
    return ()


class FalseValue(BooleanTerm):
  """Class for representing "FALSE"."""

  def simplify(self, assignments):
    return self

  def __repr__(self):
    return "FALSE"

  def __str__(self):
    return "FALSE"

  def extract_pivots(self, assignments):
    return {}

  def extract_equalities(self):
    return ()


TRUE = TrueValue()
FALSE = FalseValue()


def simplify_exprs(exprs, result_type, stop_term, skip_term):
  """Simplify a set of subexpressions for a conjunction or disjunction.

  Args:
    exprs: An iterable. The subexpressions.
    result_type: _And or _Or. The type of result (unless it simplifies
      down to something simpler).
    stop_term: FALSE for _And, TRUE for _Or. If this term is encountered,
      it will be immediately returned.
    skip_term: TRUE for _And, FALSE for _Or. If this term is encountered,
      it will be ignored.

  Returns:
    A BooleanTerm.
  """
  expr_set = set()
  for e in exprs:
    if e is stop_term:
      return stop_term
    elif e is skip_term:
      continue
    elif isinstance(e, result_type):
      expr_set = expr_set.union(e.exprs)
    else:
      expr_set.add(e)
  if len(expr_set) > 1:
    return result_type(expr_set)
  elif expr_set:
    return expr_set.pop()
  else:
    return skip_term


class _Eq(BooleanTerm):
  """An equality constraint.

  This declares an equality between a variable and a value, or a variable
  and a variable. External code should use Eq rather than creating an _Eq
  instance directly.

  Attributes:
    left: A string; left side of the equality. This is expected to be the
      string with the higher ascii value, so e.g. strings starting with "~"
      (ascii 0x7e) should be on the left.
    right: A string; right side of the equality. This is the lower ascii value.
  """
  __slots__ = ("left", "right")

  def __init__(self, left, right):
    """Initialize an equality.

    Args:
      left: A string. Left side of the equality.
      right: A string. Right side of the equality.
    """
    self.left = left
    self.right = right

  def __repr__(self):
    return f"Eq({self.left!r}, {self.right!r})"

  def __str__(self):
    return f"{self.left} == {self.right}"

  def __hash__(self):
    return hash((self.left, self.right))

  def __eq__(self, other):
    return (self.__class__ == other.__class__ and
            self.left == other.left and
            self.right == other.right)

  def __ne__(self, other):
    return not self == other

  def simplify(self, assignments):
    """Simplify this equality.

    This will try to look up the values, and return FALSE if they're no longer
    possible. Also, when comparing two variables, it will compute the
    intersection, and return a disjunction of variable=value equalities instead.

    Args:
      assignments: Variable assignments (dict mapping strings to sets of
      strings). Used to determine whether this equality is still possible, and
      to compute intersections between two variables.

    Returns:
      A new BooleanTerm.
    """
    if self.right in assignments:
      return self
    else:
      return self if self.right in assignments[self.left] else FALSE

  def extract_pivots(self, assignments):
    """Extract the pivots. See BooleanTerm.extract_pivots()."""
    if self.left in assignments and self.right in assignments:
      intersection = assignments[self.left] & assignments[self.right]
      return {self.left: frozenset(intersection),
              self.right: frozenset(intersection)}
    else:
      return {self.left: frozenset((self.right,)),
              self.right: frozenset((self.left,))}

  def extract_equalities(self):
    return ((self.left, self.right),)


def _expr_set_hash(expr_set):
  # We sort the hash of individual expressions so that two equal sets
  # have the same hash value.
  return hash(tuple(sorted(hash(e) for e in expr_set)))


class _And(BooleanTerm):
  """A conjunction of equalities and disjunctions.

  External code should use And rather than creating an _And instance directly.
  """

  __slots__ = ("exprs",)

  def __init__(self, exprs):
    """Initialize a conjunction.

    Args:
      exprs: A set. The subterms.
    """
    self.exprs = exprs

  def __eq__(self, other):
    return self.__class__ == other.__class__ and self.exprs == other.exprs

  def __ne__(self, other):
    return not self == other

  def __repr__(self):
    return f"And({list(self.exprs)!r})"

  def __str__(self):
    return "(" + " & ".join(str(t) for t in self.exprs) + ")"

  def __hash__(self):
    return _expr_set_hash(self.exprs)

  def simplify(self, assignments):
    return simplify_exprs((e.simplify(assignments) for e in self.exprs), _And,
                          FALSE, TRUE)

  def extract_pivots(self, assignments):
    """Extract the pivots. See BooleanTerm.extract_pivots()."""
    pivots = {}  # dict of frozenset
    for expr in self.exprs:
      expr_pivots = expr.extract_pivots(assignments)
      for name, values in expr_pivots.items():
        if name in pivots:
          pivots[name] = pivots[name] & values
        else:
          pivots[name] = values
    return {var: values for var, values in pivots.items() if values}

  def extract_equalities(self):
    return tuple(chain(expr.extract_equalities() for expr in self.exprs))


class _Or(BooleanTerm):
  """A disjunction of equalities and conjunctions.

  External code should use Or rather than creating an _Or instance directly.
  """

  __slots__ = ("exprs",)

  def __init__(self, exprs):
    """Initialize a disjunction.

    Args:
      exprs: A set. The subterms.
    """
    self.exprs = exprs

  def __eq__(self, other):  # for unit tests
    return self.__class__ == other.__class__ and self.exprs == other.exprs

  def __ne__(self, other):
    return not self == other

  def __repr__(self):
    return f"Or({list(self.exprs)!r})"

  def __str__(self):
    return "(" + " | ".join(str(t) for t in self.exprs) + ")"

  def __hash__(self):
    return _expr_set_hash(self.exprs)

  def simplify(self, assignments):
    return simplify_exprs((e.simplify(assignments) for e in self.exprs), _Or,
                          TRUE, FALSE)

  def extract_pivots(self, assignments):
    """Extract the pivots. See BooleanTerm.extract_pivots()."""
    pivots = {}  # dict of frozenset
    for expr in self.exprs:
      expr_pivots = expr.extract_pivots(assignments)
      for name, values in expr_pivots.items():
        if name in pivots:
          pivots[name] = pivots[name] | values
        else:
          pivots[name] = values
    return pivots

  def extract_equalities(self):
    return tuple(chain(expr.extract_equalities() for expr in self.exprs))


def Eq(left, right):  # pylint: disable=invalid-name
  """Create an equality or its simplified equivalent.

  This will ensure that left > right. (For left == right, it'll just return
  TRUE).

  Args:
    left: A string. Left side of the equality. This will get sorted, so it
      might end up on the right.
    right: A string. Right side of the equality. This will get sorted, so it
      might end up on the left.

  Returns:
    A BooleanTerm.
  """
  assert isinstance(left, str)
  assert isinstance(right, str)
  if left == right:
    return TRUE
  elif left > right:
    return _Eq(left, right)
  else:
    return _Eq(right, left)


def And(exprs):  # pylint: disable=invalid-name
  """Create a conjunction or its simplified equivalent.

  This will ensure that, when an _And is returned, none of its immediate
  subterms is TRUE, FALSE, or another conjunction.

  Args:
    exprs: An iterable. The subterms.

  Returns:
    A BooleanTerm.
  """
  return simplify_exprs(exprs, _And, FALSE, TRUE)


def Or(exprs):  # pylint: disable=invalid-name
  """Create a disjunction or its simplified equivalent.

  This will ensure that, when an _Or is returned, none of its immediate
  subterms is TRUE, FALSE, or another disjunction.

  Args:
    exprs: An iterable. The subterms.

  Returns:
    A BooleanTerm.
  """
  return simplify_exprs(exprs, _Or, TRUE, FALSE)


class Solver:
  """Solver for boolean equations.

  This solver computes the union of all solutions. I.e. rather than assigning
  exactly one value to each variable, it will create a list of values for each
  variable: All the values this variable has in any of the solutions.

  To accomplish this, we use the following rewriting rules:
    [1]  (t in X && ...) || (t in Y && ...) -->  t in (X | Y)
    [2]  t in X && t in Y                   -->  t in (X & Y)
  Applying these iteratively for each variable in turn ("extracting pivots")
  reduces the system to one where we can "read off" the possible values for each
  variable.

  Attributes:
    ANY_VALUE: A special value assigned to variables with no constraints.
    variables: A list of all variables.
    values: A list of all values.
    implications: A nested dictionary mapping variable names to values to
      BooleanTerm instances. This is used to specify rules like "if x is 1,
      then ..."
    ground_truths: An equation that needs to always be TRUE. If this is FALSE,
      or can be reduced to FALSE, the system is unsolvable.
  """

  ANY_VALUE = "?"

  def __init__(self):
    self.variables = set()
    self.implications = collections.defaultdict(dict)
    self.ground_truth = TRUE
    self.assignments = None

  def __str__(self):
    lines = []
    count_false, count_true = 0, 0
    if self.ground_truth is not TRUE:
      lines.append(f"always: {self.ground_truth}")
    for var, value, implication in self._iter_implications():
      # only print the "interesting" lines
      if implication is FALSE:
        count_false += 1
      elif implication is TRUE:
        count_true += 1
      else:
        lines.append(f"if {_Eq(var, value)} then {implication}")
    return "%s\n(not shown: %d always FALSE, %d always TRUE)\n" % (
        "\n".join(lines), count_false, count_true)

  def __repr__(self):
    lines = []
    for var in self.variables:
      lines.append(f"solver.register_variable({var!r})")
    if self.ground_truth is not TRUE:
      lines.append(f"solver.always_true({self.ground_truth!r})")
    for var, value, implication in self._iter_implications():
      lines.append(f"solver.implies({_Eq(var, value)!r}, {implication!r})")
    return "\n" + "".join(line + "\n" for line in lines)

  def register_variable(self, variable):
    """Register a variable. Call before calling solve()."""
    self.variables.add(variable)

  def always_true(self, formula):
    """Register a ground truth. Call before calling solve()."""
    assert formula is not FALSE
    self.ground_truth = And([self.ground_truth, formula])

  def implies(self, e, implication):
    """Register an implication. Call before calling solve()."""
    # COV_NF_START
    if e is FALSE or e is TRUE:
      raise AssertionError("Illegal equation")
    # COV_NF_END
    assert isinstance(e, _Eq)
    assert e.right not in self.implications[e.left]
    # Since _Eq sorts its arguments in reverse and variables start with "~"
    # (ASCII value 126), e.left should always be the variable.
    self.implications[e.left][e.right] = implication

  def _iter_implications(self):
    for var, value_to_implication in self.implications.items():
      for value, implication in value_to_implication.items():
        yield (var, value, implication)

  def _get_nonfalse_values(self, var):
    return {value for value, implication in self.implications[var].items()
            if implication is not FALSE}

  def _get_first_approximation(self):
    """Get all (variable, value) combinations to consider.

    This gets the (variable, value) combinations that the solver needs to
    consider based on the equalities that appear in the implications. E.g.,
    with the following implication:
      t1 = v1 => t1 = t2 | t3 = v2
    the combinations to consider are
      (t1, v1) because t1 = v1 appears,
      (t2, v1) because t1 = t2 and t1 = v1 appear, and
      (t3, v2) because t3 = v2 appears.

    Returns:
      A dictionary D mapping strings (variables) to sets of strings
      (values). For two variables t1 and t2, if t1 = t2 is a possible
      assignment (by first approximation), then D[t1] and D[t2] point
      to the same memory location.
    """
    equalities = set(chain(
        implication.extract_equalities()
        for (_, _, implication) in self._iter_implications())).union(
            self.ground_truth.extract_equalities())
    var_assignments = {}
    value_assignments = {}
    for var in self.variables:
      var_assignments[var] = {var}
      value_assignments[var] = self._get_nonfalse_values(var)

    for var, value in equalities:
      if value in self.variables:
        other_var = value
        value_assignments[var] |= value_assignments[other_var]
        for var_assignment in var_assignments[other_var]:
          var_assignments[var].add(var_assignment)
          # Make the two variables point to the same sets of assignments so
          # that further possible assignments for either are added to both.
          var_assignments[var_assignment] = var_assignments[var]
          value_assignments[var_assignment] = value_assignments[var]
      else:
        value_assignments[var].add(value)

    return value_assignments

  def _complete(self):
    """Insert missing implications.

    Insert all implications needed to have one implication for every
    (variable, value) combination returned by _get_first_approximation().
    """
    for var, values in self._get_first_approximation().items():
      for value in values:
        if value not in self.implications[var]:
          # Missing implications are typically needed for variable/value
          # combinations not considered by the user, e.g. for auxiliary
          # variables introduced when setting up the "main" equations.
          self.implications[var][value] = TRUE
      if not self.implications[var]:
        # If a variable does not have any constraints, it can be anything.
        self.implications[var][Solver.ANY_VALUE] = TRUE

  def solve(self):
    """Solve the system of equations.

    Returns:
      An assignment, mapping strings (variables) to sets of strings (values).
    """
    if self.assignments:
      return self.assignments

    self._complete()

    assignments = {var: self._get_nonfalse_values(var)
                   for var in self.variables}

    ground_pivots = self.ground_truth.simplify(assignments).extract_pivots(
        assignments)
    for pivot, possible_values in ground_pivots.items():
      if pivot in assignments:
        assignments[pivot] &= set(possible_values)

    something_changed = True
    while something_changed:
      something_changed = False

      and_terms = []
      for var in self.variables:
        or_terms = []
        for value in assignments[var].copy():
          implication = self.implications[var][value].simplify(assignments)
          if implication is FALSE:
            # As an example of what kind of code triggers this,
            # see TestBoolEq.testFilter
            assignments[var].remove(value)
            something_changed = True
          else:
            or_terms.append(implication)
          self.implications[var][value] = implication
        and_terms.append(Or(or_terms))
      d = And(and_terms)

      for pivot, possible_values in d.extract_pivots(assignments).items():
        if pivot in assignments:
          length_before = len(assignments[pivot])
          assignments[pivot] &= set(possible_values)
          length_after = len(assignments[pivot])
          something_changed |= (length_before != length_after)

    self.register_variable = pytd_utils.disabled_function
    self.implies = pytd_utils.disabled_function

    self.assignments = assignments
    return assignments



# flake8-2020 has type annotations, but PEP 561 states:
# This PEP does not support distributing typing information as part of module-only distributions or single-file modules within namespace packages.
# Therefore typeshed is the best place.

import ast
from typing import Any, ClassVar, Generator

class Plugin:
    name: ClassVar[str]
    version: ClassVar[str]
    def __init__(self, tree: ast.AST) -> None: ...
    def run(self) -> Generator[tuple[int, int, str, type[Any]], None, None]: ...

def __getattr__(name: str) -> Any: ...  # incomplete (other attributes are normally not accessed)



"""Inspect the CFG and typegraph."""

import logging
import subprocess

import networkx as nx


log = logging.getLogger(__name__)


def obj_key(n):
  nid = n.id if hasattr(n, "id") else id(n)
  return n.__class__.__name__ + str(nid)


def obj_repr(n):
  return repr(n.data)[:10]


class TypeGraph:
  """Networkx graph builder."""

  def __init__(self, program, ignored, only_cfg=False):
    self.graph = nx.MultiDiGraph()
    self._add_cfg(program, ignored)
    if not only_cfg:
      self._add_variables(program, ignored)

  def add_node(self, obj, **kwargs):
    self.graph.add_node(obj_key(obj), **kwargs)

  def add_edge(self, obj1, obj2, **kwargs):
    self.graph.add_edge(obj_key(obj1), obj_key(obj2), **kwargs)

  def to_dot(self):
    return nx.nx_pydot.to_pydot(self.graph).to_string()

  def _add_cfg(self, program, ignored):
    """Add program cfg nodes."""

    for node in program.cfg_nodes:
      if node in ignored:
        continue
      self.add_node(
          node, label=f"<{node.id}>{node.name}", shape="polygon", sides=4)
      for other in node.outgoing:
        self.add_edge(node, other, penwidth=2.0)

  def _add_variables(self, program, ignored):
    """A dd program variables and bindings."""

    def _is_constant(val):
      return all(origin.where == program.entrypoint for origin in val.origins)

    for variable in program.variables:
      if variable.id in ignored:
        continue
      # Ignore "boring" values (a.k.a. constants)
      if all(_is_constant(value) for value in variable.bindings):
        continue
      self.add_node(
          variable, label=f"v{variable.id}", shape="polygon", sides=4,
          distortion=.1)
      for val in variable.bindings:
        label = f"{obj_repr(val)}@0x{id(val.data):x}"
        color = "white" if val.origins else "red"
        self.add_node(val, label=label, fillcolor=color)
        self.add_edge(variable, val, arrowhead="none")
        for origin in val.origins:
          if origin.where == program.entrypoint:
            continue
          for srcs in origin.source_sets:
            self.add_node(srcs, label="")
            self.add_edge(val, srcs, color="pink", arrowhead="none", weight=40)
            if origin.where not in ignored:
              self.add_edge(origin.where, srcs, arrowhead="none",
                            style="dotted", weight=5)
            for src in srcs:
              self.add_edge(src, srcs, color="lightblue", weight=2)


def write_svg_from_dot(svg_file, dot):
  with subprocess.Popen(
      ["/usr/bin/dot", "-T", "svg", "-o", svg_file],
      stdin=subprocess.PIPE, universal_newlines=True) as proc:
    (_, stderr) = proc.communicate(dot)
  if stderr:
    log.info("Failed to create %s: %s", svg_file, stderr)



from typing import Any, Pattern, Text

# rx can be any object with a 'search' method; once we have Protocols we can change the type
def compile_dir(
    dir: Text, maxlevels: int = ..., ddir: Text | None = ..., force: bool = ..., rx: Pattern[Any] | None = ..., quiet: int = ...
) -> int: ...
def compile_file(
    fullname: Text, ddir: Text | None = ..., force: bool = ..., rx: Pattern[Any] | None = ..., quiet: int = ...
) -> int: ...
def compile_path(skip_curdir: bool = ..., maxlevels: int = ..., force: bool = ..., quiet: int = ...) -> int: ...



from typing import Any

def instances(cursor, context): ...
def merge_frozen_result(session, statement, frozen_result, load: bool = ...): ...
def merge_result(query, iterator, load: bool = ...): ...
def get_from_identity(session, mapper, key, passive): ...
def load_on_ident(
    session,
    statement,
    key,
    load_options: Any | None = ...,
    refresh_state: Any | None = ...,
    with_for_update: Any | None = ...,
    only_load_props: Any | None = ...,
    no_autoflush: bool = ...,
    bind_arguments=...,
    execution_options=...,
): ...
def load_on_pk_identity(
    session,
    statement,
    primary_key_identity,
    load_options: Any | None = ...,
    refresh_state: Any | None = ...,
    with_for_update: Any | None = ...,
    only_load_props: Any | None = ...,
    identity_token: Any | None = ...,
    no_autoflush: bool = ...,
    bind_arguments=...,
    execution_options=...,
): ...

class PostLoad:
    loaders: Any
    states: Any
    load_keys: Any
    def __init__(self) -> None: ...
    def add_state(self, state, overwrite) -> None: ...
    def invoke(self, context, path) -> None: ...
    @classmethod
    def for_context(cls, context, path, only_load_props): ...
    @classmethod
    def path_exists(cls, context, path, key): ...
    @classmethod
    def callable_for_path(cls, context, path, limit_to_mapper, token, loader_callable, *arg, **kw) -> None: ...

def load_scalar_attributes(mapper, state, attribute_names, passive) -> None: ...



from . import _LangModelType

Latin7_char_to_order_map: tuple[int, ...]
win1253_char_to_order_map: tuple[int, ...]
GreekLangModel: tuple[int, ...]
Latin7GreekModel: _LangModelType
Win1253GreekModel: _LangModelType



from typing import Any

from openpyxl.descriptors.serialisable import Serialisable

from .cell_range import CellRange

class MergeCell(CellRange):
    tagname: str
    @property
    def ref(self): ...
    __attrs__: Any
    def __init__(self, ref: Any | None = ...) -> None: ...
    def __copy__(self): ...

class MergeCells(Serialisable):
    tagname: str
    # Overwritten by property below
    # count: Integer
    mergeCell: Any
    __elements__: Any
    __attrs__: Any
    def __init__(self, count: Any | None = ..., mergeCell=...) -> None: ...
    @property
    def count(self): ...

class MergedCellRange(CellRange):
    ws: Any
    start_cell: Any
    def __init__(self, worksheet, coord) -> None: ...
    def format(self) -> None: ...
    def __contains__(self, coord): ...
    def __copy__(self): ...



import sys
from _typeshed import Self, StrOrBytesPath
from datetime import date, datetime, time
from types import TracebackType
from typing import Any, Callable, Generator, Iterable, Iterator, Protocol, TypeVar
from typing_extensions import Literal, final

_T = TypeVar("_T")

paramstyle: str
threadsafety: int
apilevel: str
Date = date
Time = time
Timestamp = datetime

def DateFromTicks(ticks: float) -> Date: ...
def TimeFromTicks(ticks: float) -> Time: ...
def TimestampFromTicks(ticks: float) -> Timestamp: ...

version_info: tuple[int, int, int]
sqlite_version_info: tuple[int, int, int]
Binary = memoryview

# The remaining definitions are imported from _sqlite3.

PARSE_COLNAMES: int
PARSE_DECLTYPES: int
SQLITE_ALTER_TABLE: int
SQLITE_ANALYZE: int
SQLITE_ATTACH: int
SQLITE_CREATE_INDEX: int
SQLITE_CREATE_TABLE: int
SQLITE_CREATE_TEMP_INDEX: int
SQLITE_CREATE_TEMP_TABLE: int
SQLITE_CREATE_TEMP_TRIGGER: int
SQLITE_CREATE_TEMP_VIEW: int
SQLITE_CREATE_TRIGGER: int
SQLITE_CREATE_VIEW: int
if sys.version_info >= (3, 7):
    SQLITE_CREATE_VTABLE: int
SQLITE_DELETE: int
SQLITE_DENY: int
SQLITE_DETACH: int
if sys.version_info >= (3, 7):
    SQLITE_DONE: int
SQLITE_DROP_INDEX: int
SQLITE_DROP_TABLE: int
SQLITE_DROP_TEMP_INDEX: int
SQLITE_DROP_TEMP_TABLE: int
SQLITE_DROP_TEMP_TRIGGER: int
SQLITE_DROP_TEMP_VIEW: int
SQLITE_DROP_TRIGGER: int
SQLITE_DROP_VIEW: int
if sys.version_info >= (3, 7):
    SQLITE_DROP_VTABLE: int
    SQLITE_FUNCTION: int
SQLITE_IGNORE: int
SQLITE_INSERT: int
SQLITE_OK: int
if sys.version_info >= (3, 11):
    SQLITE_LIMIT_LENGTH: int
    SQLITE_LIMIT_SQL_LENGTH: int
    SQLITE_LIMIT_COLUMN: int
    SQLITE_LIMIT_EXPR_DEPTH: int
    SQLITE_LIMIT_COMPOUND_SELECT: int
    SQLITE_LIMIT_VDBE_OP: int
    SQLITE_LIMIT_FUNCTION_ARG: int
    SQLITE_LIMIT_ATTACHED: int
    SQLITE_LIMIT_LIKE_PATTERN_LENGTH: int
    SQLITE_LIMIT_VARIABLE_NUMBER: int
    SQLITE_LIMIT_TRIGGER_DEPTH: int
    SQLITE_LIMIT_WORKER_THREADS: int
SQLITE_PRAGMA: int
SQLITE_READ: int
SQLITE_REINDEX: int
if sys.version_info >= (3, 7):
    SQLITE_RECURSIVE: int
    SQLITE_SAVEPOINT: int
SQLITE_SELECT: int
SQLITE_TRANSACTION: int
SQLITE_UPDATE: int
adapters: Any
converters: Any
sqlite_version: str
version: str

# TODO: adapt needs to get probed
def adapt(obj, protocol, alternate): ...
def complete_statement(statement: str) -> bool: ...

if sys.version_info >= (3, 7):
    def connect(
        database: StrOrBytesPath,
        timeout: float = ...,
        detect_types: int = ...,
        isolation_level: str | None = ...,
        check_same_thread: bool = ...,
        factory: type[Connection] | None = ...,
        cached_statements: int = ...,
        uri: bool = ...,
    ) -> Connection: ...

else:
    def connect(
        database: bytes | str,
        timeout: float = ...,
        detect_types: int = ...,
        isolation_level: str | None = ...,
        check_same_thread: bool = ...,
        factory: type[Connection] | None = ...,
        cached_statements: int = ...,
        uri: bool = ...,
    ) -> Connection: ...

def enable_callback_tracebacks(__enable: bool) -> None: ...
def enable_shared_cache(enable: int) -> None: ...
def register_adapter(__type: type[_T], __caster: Callable[[_T], int | float | str | bytes]) -> None: ...
def register_converter(__name: str, __converter: Callable[[bytes], Any]) -> None: ...

if sys.version_info < (3, 8):
    class Cache:
        def __init__(self, *args, **kwargs) -> None: ...
        def display(self, *args, **kwargs) -> None: ...
        def get(self, *args, **kwargs) -> None: ...

class _AggregateProtocol(Protocol):
    def step(self, value: int) -> None: ...
    def finalize(self) -> int: ...

class Connection:
    DataError: Any
    DatabaseError: Any
    Error: Any
    IntegrityError: Any
    InterfaceError: Any
    InternalError: Any
    NotSupportedError: Any
    OperationalError: Any
    ProgrammingError: Any
    Warning: Any
    in_transaction: Any
    isolation_level: Any
    row_factory: Any
    text_factory: Any
    total_changes: Any
    def __init__(self, *args: Any, **kwargs: Any) -> None: ...
    def close(self) -> None: ...
    def commit(self) -> None: ...
    def create_aggregate(self, name: str, n_arg: int, aggregate_class: Callable[[], _AggregateProtocol]) -> None: ...
    def create_collation(self, __name: str, __callback: Any) -> None: ...
    if sys.version_info >= (3, 8):
        def create_function(self, name: str, narg: int, func: Any, *, deterministic: bool = ...) -> None: ...
    else:
        def create_function(self, name: str, num_params: int, func: Any) -> None: ...

    def cursor(self, cursorClass: type | None = ...) -> Cursor: ...
    def execute(self, sql: str, parameters: Iterable[Any] = ...) -> Cursor: ...
    # TODO: please check in executemany() if seq_of_parameters type is possible like this
    def executemany(self, __sql: str, __parameters: Iterable[Iterable[Any]]) -> Cursor: ...
    def executescript(self, __sql_script: bytes | str) -> Cursor: ...
    def interrupt(self) -> None: ...
    def iterdump(self) -> Generator[str, None, None]: ...
    def rollback(self) -> None: ...
    def set_authorizer(
        self, authorizer_callback: Callable[[int, str | None, str | None, str | None, str | None], int] | None
    ) -> None: ...
    def set_progress_handler(self, progress_handler: Callable[[], bool | None] | None, n: int) -> None: ...
    def set_trace_callback(self, trace_callback: Callable[[str], object] | None) -> None: ...
    # enable_load_extension and load_extension is not available on python distributions compiled
    # without sqlite3 loadable extension support. see footnotes https://docs.python.org/3/library/sqlite3.html#f1
    def enable_load_extension(self, enabled: bool) -> None: ...
    def load_extension(self, path: str) -> None: ...
    if sys.version_info >= (3, 7):
        def backup(
            self,
            target: Connection,
            *,
            pages: int = ...,
            progress: Callable[[int, int, int], object] | None = ...,
            name: str = ...,
            sleep: float = ...,
        ) -> None: ...

    def __call__(self, *args: Any, **kwargs: Any) -> Any: ...
    def __enter__(self: Self) -> Self: ...
    def __exit__(
        self, __type: type[BaseException] | None, __value: BaseException | None, __traceback: TracebackType | None
    ) -> Literal[False]: ...

class Cursor(Iterator[Any]):
    arraysize: Any
    connection: Any
    description: Any
    lastrowid: Any
    row_factory: Any
    rowcount: int
    # TODO: Cursor class accepts exactly 1 argument
    # required type is sqlite3.Connection (which is imported as _Connection)
    # however, the name of the __init__ variable is unknown
    def __init__(self, *args: Any, **kwargs: Any) -> None: ...
    def close(self) -> None: ...
    def execute(self, __sql: str, __parameters: Iterable[Any] = ...) -> Cursor: ...
    def executemany(self, __sql: str, __seq_of_parameters: Iterable[Iterable[Any]]) -> Cursor: ...
    def executescript(self, __sql_script: bytes | str) -> Cursor: ...
    def fetchall(self) -> list[Any]: ...
    def fetchmany(self, size: int | None = ...) -> list[Any]: ...
    def fetchone(self) -> Any: ...
    def setinputsizes(self, __sizes: object) -> None: ...  # does nothing
    def setoutputsize(self, __size: object, __column: object = ...) -> None: ...  # does nothing
    def __iter__(self: Self) -> Self: ...
    def __next__(self) -> Any: ...

class DataError(DatabaseError): ...
class DatabaseError(Error): ...

class Error(Exception):
    if sys.version_info >= (3, 11):
        sqlite_errorcode: int
        sqlite_errorname: str

class IntegrityError(DatabaseError): ...
class InterfaceError(Error): ...
class InternalError(DatabaseError): ...
class NotSupportedError(DatabaseError): ...
class OperationalError(DatabaseError): ...

OptimizedUnicode = str

@final
class PrepareProtocol:
    def __init__(self, *args: Any, **kwargs: Any) -> None: ...

class ProgrammingError(DatabaseError): ...

class Row:
    def __init__(self, *args: Any, **kwargs: Any) -> None: ...
    def keys(self): ...
    def __eq__(self, __other): ...
    def __ge__(self, __other): ...
    def __getitem__(self, __index): ...
    def __gt__(self, __other): ...
    def __hash__(self): ...
    def __iter__(self): ...
    def __le__(self, __other): ...
    def __len__(self): ...
    def __lt__(self, __other): ...
    def __ne__(self, __other): ...

if sys.version_info < (3, 8):
    @final
    class Statement:
        def __init__(self, *args, **kwargs): ...

class Warning(Exception): ...



from typing import Any, AnyStr, Callable, Iterator, Match, Pattern, overload

# ----- re variables and constants -----
DEBUG: int
I: int
IGNORECASE: int
L: int
LOCALE: int
M: int
MULTILINE: int
S: int
DOTALL: int
X: int
VERBOSE: int
U: int
UNICODE: int
T: int
TEMPLATE: int

class error(Exception): ...

@overload
def compile(pattern: AnyStr, flags: int = ...) -> Pattern[AnyStr]: ...
@overload
def compile(pattern: Pattern[AnyStr], flags: int = ...) -> Pattern[AnyStr]: ...
@overload
def search(pattern: str | unicode, string: AnyStr, flags: int = ...) -> Match[AnyStr] | None: ...
@overload
def search(pattern: Pattern[str] | Pattern[unicode], string: AnyStr, flags: int = ...) -> Match[AnyStr] | None: ...
@overload
def match(pattern: str | unicode, string: AnyStr, flags: int = ...) -> Match[AnyStr] | None: ...
@overload
def match(pattern: Pattern[str] | Pattern[unicode], string: AnyStr, flags: int = ...) -> Match[AnyStr] | None: ...
@overload
def split(pattern: str | unicode, string: AnyStr, maxsplit: int = ..., flags: int = ...) -> list[AnyStr]: ...
@overload
def split(pattern: Pattern[str] | Pattern[unicode], string: AnyStr, maxsplit: int = ..., flags: int = ...) -> list[AnyStr]: ...
@overload
def findall(pattern: str | unicode, string: AnyStr, flags: int = ...) -> list[Any]: ...
@overload
def findall(pattern: Pattern[str] | Pattern[unicode], string: AnyStr, flags: int = ...) -> list[Any]: ...

# Return an iterator yielding match objects over all non-overlapping matches
# for the RE pattern in string. The string is scanned left-to-right, and
# matches are returned in the order found. Empty matches are included in the
# result unless they touch the beginning of another match.
@overload
def finditer(pattern: str | unicode, string: AnyStr, flags: int = ...) -> Iterator[Match[AnyStr]]: ...
@overload
def finditer(pattern: Pattern[str] | Pattern[unicode], string: AnyStr, flags: int = ...) -> Iterator[Match[AnyStr]]: ...
@overload
def sub(pattern: str | unicode, repl: AnyStr, string: AnyStr, count: int = ..., flags: int = ...) -> AnyStr: ...
@overload
def sub(
    pattern: str | unicode, repl: Callable[[Match[AnyStr]], AnyStr], string: AnyStr, count: int = ..., flags: int = ...
) -> AnyStr: ...
@overload
def sub(pattern: Pattern[str] | Pattern[unicode], repl: AnyStr, string: AnyStr, count: int = ..., flags: int = ...) -> AnyStr: ...
@overload
def sub(
    pattern: Pattern[str] | Pattern[unicode],
    repl: Callable[[Match[AnyStr]], AnyStr],
    string: AnyStr,
    count: int = ...,
    flags: int = ...,
) -> AnyStr: ...
@overload
def subn(pattern: str | unicode, repl: AnyStr, string: AnyStr, count: int = ..., flags: int = ...) -> tuple[AnyStr, int]: ...
@overload
def subn(
    pattern: str | unicode, repl: Callable[[Match[AnyStr]], AnyStr], string: AnyStr, count: int = ..., flags: int = ...
) -> tuple[AnyStr, int]: ...
@overload
def subn(
    pattern: Pattern[str] | Pattern[unicode], repl: AnyStr, string: AnyStr, count: int = ..., flags: int = ...
) -> tuple[AnyStr, int]: ...
@overload
def subn(
    pattern: Pattern[str] | Pattern[unicode],
    repl: Callable[[Match[AnyStr]], AnyStr],
    string: AnyStr,
    count: int = ...,
    flags: int = ...,
) -> tuple[AnyStr, int]: ...
def escape(string: AnyStr) -> AnyStr: ...
def purge() -> None: ...
def template(pattern: AnyStr | Pattern[AnyStr], flags: int = ...) -> Pattern[AnyStr]: ...



"""Tools for output generation."""

import collections
import contextlib
import enum
import logging
from typing import cast

from pytype import special_builtins
from pytype import utils
from pytype.abstract import abstract
from pytype.abstract import abstract_utils
from pytype.abstract import class_mixin
from pytype.abstract import function
from pytype.overlays import attr_overlay
from pytype.overlays import dataclass_overlay
from pytype.overlays import named_tuple
from pytype.overlays import typed_dict
from pytype.overlays import typing_overlay
from pytype.pyi import metadata
from pytype.pytd import pytd
from pytype.pytd import pytd_utils
from pytype.pytd import visitors

log = logging.getLogger(__name__)


class Converter(utils.ContextWeakrefMixin):
  """Functions for converting abstract classes into PyTD."""

  class OutputMode(enum.IntEnum):
    """Controls the level of detail in pytd types. See set_output_mode."""
    NORMAL = 0
    DETAILED = 1
    LITERAL = 2

  def __init__(self, ctx):
    super().__init__(ctx)
    self._output_mode = Converter.OutputMode.NORMAL
    self._scopes = []

  @contextlib.contextmanager
  def set_output_mode(self, mode):
    """Change the level of detail in pytd types.

    Args:
      mode: Converter.OutputMode option controlling the level of detail to use.
        NORMAL - the default, safe for pyi files.
        DETAILED - more detail, unsafe for pyi files. The converter will do
          things like using the names of inner classes rather than Any and
          including the known argument types for a callable even if the argument
          count is unknown. Useful for error messages.
        LITERAL - like DETAILED, but bool, int, str, and bytes constants will be
          emitted as Literal[<constant>] rather than their type.

    Yields:
      None.
    """
    old = self._output_mode
    self._output_mode = mode
    yield
    self._output_mode = old

  @property
  def _detailed(self):
    return self._output_mode >= Converter.OutputMode.DETAILED

  def _get_values(self, node, var, view):
    if var.bindings and view is not None:
      return [view[var].data]
    elif node:
      return var.FilteredData(node, strict=False)
    else:
      return var.data

  def _is_tuple(self, v, instance):
    return (isinstance(v, abstract.TupleClass) or
            isinstance(instance, abstract.Tuple))

  def _value_to_parameter_types(self, node, v, instance, template, seen, view):
    """Get PyTD types for the parameters of an instance of an abstract value."""
    if isinstance(v, abstract.CallableClass):
      assert template == (abstract_utils.ARGS, abstract_utils.RET), template
      template = list(range(v.num_args)) + [template[1]]
    if self._is_tuple(v, instance):
      if isinstance(v, abstract.TupleClass):
        new_template = range(v.tuple_length)
      else:
        new_template = range(instance.tuple_length)
      if template:
        assert len(template) == 1 and template[0] == abstract_utils.T, template
      else:
        # We have a recursive type. By erasing the instance and value
        # information, we'll return Any for all of the tuple elements.
        v = instance = None
      template = new_template
    if instance is None and isinstance(v, abstract.ParameterizedClass):
      assert v
      return [self.value_instance_to_pytd_type(
          node, v.get_formal_type_parameter(t), None, seen, view)
              for t in template]
    elif isinstance(instance, abstract.SimpleValue):
      assert instance
      type_arguments = []
      for t in template:
        if isinstance(instance, abstract.Tuple):
          param_values = {val: view for val in self._get_values(
              node, instance.pyval[t], view)}
        elif instance.has_instance_type_parameter(t):
          param_values = {val: view for val in self._get_values(
              node, instance.get_instance_type_parameter(t), view)}
        elif isinstance(v, abstract.CallableClass):
          param_node = node or self.ctx.root_node
          param_var = v.get_formal_type_parameter(t).instantiate(param_node)
          if view is None:
            param_values = {val: None for val in param_var.data}
          else:
            param_values = {}
            for new_view in abstract_utils.get_views([param_var], param_node):
              new_view.update(view)
              param_values[new_view[param_var].data] = new_view
        else:
          param_values = {self.ctx.convert.unsolvable: view}
        formal_param = v.get_formal_type_parameter(t)
        # If the instance's parameter value is unsolvable or the parameter type
        # is recursive, we can get a more precise type from the class. Note that
        # we need to be careful not to introduce unbound type parameters.
        if (isinstance(v, abstract.ParameterizedClass) and
            not formal_param.formal and
            (list(param_values.keys()) == [self.ctx.convert.unsolvable] or
             abstract_utils.is_recursive_annotation(formal_param))):
          arg = self.value_instance_to_pytd_type(
              node, formal_param, None, seen, view)
        else:
          arg = pytd_utils.JoinTypes(
              self.value_to_pytd_type(node, p, seen, param_view)
              for p, param_view in param_values.items())
        type_arguments.append(arg)
      return type_arguments
    else:
      return [pytd.AnythingType() for _ in template]

  def value_instance_to_pytd_type(self, node, v, instance, seen, view):
    """Get the PyTD type an instance of this object would have.

    Args:
      node: The node.
      v: The object.
      instance: The instance.
      seen: Already seen instances.
      view: A Variable -> binding map.

    Returns:
      A PyTD type.
    """
    if abstract_utils.is_recursive_annotation(v):
      return pytd.LateType(v.unflatten_expr() if self._detailed else v.expr)
    elif isinstance(v, abstract.Union):
      return pytd.UnionType(tuple(
          self.value_instance_to_pytd_type(node, t, instance, seen, view)
          for t in v.options))
    elif isinstance(v, abstract.AnnotationContainer):
      return self.value_instance_to_pytd_type(
          node, v.base_cls, instance, seen, view)
    elif isinstance(v, abstract.LiteralClass):
      if isinstance(v.value, abstract.Instance) and v.value.cls.is_enum:
        typ = pytd_utils.NamedTypeWithModule(
            v.value.cls.official_name or v.value.cls.name, v.value.cls.module)
        value = pytd.Constant(v.value.name, typ)
      elif isinstance(v.value.pyval, (str, bytes)):
        # Strings are stored as strings of their representations, prefix and
        # quotes and all.
        value = repr(v.value.pyval)
      elif isinstance(v.value.pyval, bool):
        # True and False are stored as pytd constants.
        value = self.ctx.loader.lookup_builtin(f"builtins.{v.value.pyval}")
      else:
        # Ints are stored as their literal values. Note that Literal[None] or a
        # nested literal will never appear here, since we simplified it to None
        # or unnested it, respectively, in typing_overlay.
        assert isinstance(v.value.pyval, int), v.value.pyval
        value = v.value.pyval
      return pytd.Literal(value)
    elif isinstance(v, typed_dict.TypedDictClass):
      # TypedDict inherits from abstract.Dict for analysis purposes, but when
      # outputting to a pyi we do not want to treat it as a generic type.
      return pytd.NamedType(v.name)
    elif isinstance(v, abstract.Class):
      if not self._detailed and v.official_name is None:
        return pytd.AnythingType()
      if seen is None:
        # We make the set immutable to ensure that the seen instances for
        # different parameter values don't interfere with one another.
        seen = frozenset()
      if instance in seen:
        # We have a circular dependency in our types (e.g., lst[0] == lst). Stop
        # descending into the type parameters.
        type_params = ()
      else:
        type_params = tuple(t.name for t in v.template)
      if instance is not None:
        seen |= {instance}
      type_arguments = self._value_to_parameter_types(
          node, v, instance, type_params, seen, view)
      base = pytd_utils.NamedTypeWithModule(v.official_name or v.name, v.module)
      if self._is_tuple(v, instance):
        homogeneous = False
      elif v.full_name == "typing.Callable":
        homogeneous = not isinstance(v, abstract.CallableClass)
      else:
        homogeneous = len(type_arguments) == 1
      return pytd_utils.MakeClassOrContainerType(
          base, type_arguments, homogeneous)
    elif isinstance(v, abstract.TypeParameter):
      # We generate the full definition because, if this type parameter is
      # imported, we will need the definition in order to declare it later.
      return self._typeparam_to_def(node, v, v.name)
    elif isinstance(v, typing_overlay.NoReturn):
      return pytd.NothingType()
    else:
      log.info("Using Any for instance of %s", v.name)
      return pytd.AnythingType()

  def value_to_pytd_type(self, node, v, seen, view):
    """Get a PyTD type representing this object, as seen at a node.

    Args:
      node: The node from which we want to observe this object.
      v: The object.
      seen: The set of values seen before while computing the type.
      view: A Variable -> binding map.

    Returns:
      A PyTD type.
    """
    if isinstance(v, (abstract.Empty, typing_overlay.NoReturn)):
      return pytd.NothingType()
    elif isinstance(v, abstract.TypeParameterInstance):
      if (v.module in self._scopes or
          v.instance is abstract_utils.DUMMY_CONTAINER):
        return self._typeparam_to_def(node, v.param, v.param.name)
      elif v.instance.get_instance_type_parameter(v.full_name).bindings:
        # The type parameter was initialized. Set the view to None, since we
        # don't include v.instance in the view.
        return pytd_utils.JoinTypes(
            self.value_to_pytd_type(node, p, seen, None)
            for p in v.instance.get_instance_type_parameter(v.full_name).data)
      elif v.param.constraints:
        return pytd_utils.JoinTypes(
            self.value_instance_to_pytd_type(node, p, None, seen, view)
            for p in v.param.constraints)
      elif v.param.bound:
        return self.value_instance_to_pytd_type(
            node, v.param.bound, None, seen, view)
      else:
        return pytd.AnythingType()
    elif isinstance(v, typing_overlay.TypeVar):
      return pytd.NamedType("builtins.type")
    elif isinstance(v, dataclass_overlay.FieldInstance):
      if not v.default:
        return pytd.AnythingType()
      return pytd_utils.JoinTypes(
          self.value_to_pytd_type(node, d, seen, view)
          for d in v.default.data)
    elif isinstance(v, attr_overlay.AttribInstance):
      ret = self.value_to_pytd_type(node, v.typ, seen, view)
      md = metadata.to_pytd(v.to_metadata())
      return pytd.Annotated(ret, ("'pytype_metadata'", md))
    elif isinstance(v, special_builtins.PropertyInstance):
      return pytd.NamedType("builtins.property")
    elif isinstance(v, typed_dict.TypedDict):
      return pytd.NamedType(v.props.name)
    elif isinstance(v, abstract.FUNCTION_TYPES):
      try:
        signatures = function.get_signatures(v)
      except NotImplementedError:
        return pytd.NamedType("typing.Callable")
      if len(signatures) == 1:
        val = self.signature_to_callable(signatures[0])
        if not isinstance(v, abstract.PYTD_FUNCTION_TYPES) or not val.formal:
          # This is a workaround to make sure we don't put unexpected type
          # parameters in call traces.
          return self.value_instance_to_pytd_type(node, val, None, seen, view)
      return pytd.NamedType("typing.Callable")
    elif isinstance(v, (abstract.ClassMethod, abstract.StaticMethod)):
      return self.value_to_pytd_type(node, v.method, seen, view)
    elif isinstance(v, (special_builtins.IsInstance,
                        special_builtins.ClassMethodCallable)):
      return pytd.NamedType("typing.Callable")
    elif isinstance(v, abstract.Class):
      param = self.value_instance_to_pytd_type(node, v, None, seen, view)
      return pytd.GenericType(base_type=pytd.NamedType("builtins.type"),
                              parameters=(param,))
    elif isinstance(v, abstract.Module):
      return pytd.Alias(v.name, pytd.Module(v.name, module_name=v.full_name))
    elif (self._output_mode >= Converter.OutputMode.LITERAL and
          isinstance(v, abstract.ConcreteValue) and
          isinstance(v.pyval, (int, str, bytes))):
      # LITERAL mode is used only for pretty-printing, so we just stringify the
      # inner value rather than properly converting it.
      return pytd.Literal(repr(v.pyval))
    elif isinstance(v, abstract.SimpleValue):
      ret = self.value_instance_to_pytd_type(
          node, v.cls, v, seen=seen, view=view)
      ret.Visit(
          visitors.FillInLocalPointers({"builtins": self.ctx.loader.builtins}))
      return ret
    elif isinstance(v, abstract.Union):
      return pytd_utils.JoinTypes(self.value_to_pytd_type(node, o, seen, view)
                                  for o in v.options)
    elif isinstance(v, special_builtins.SuperInstance):
      return pytd.NamedType("builtins.super")
    elif isinstance(v, abstract.TypeParameter):
      # Arguably, the type of a type parameter is NamedType("typing.TypeVar"),
      # but pytype doesn't know how to handle that, so let's just go with Any
      # unless self._detailed is set.
      if self._detailed:
        return pytd.NamedType("typing.TypeVar")
      else:
        return pytd.AnythingType()
    elif isinstance(v, abstract.Unsolvable):
      return pytd.AnythingType()
    elif isinstance(v, abstract.Unknown):
      return pytd.NamedType(v.class_name)
    elif isinstance(v, abstract.BuildClass):
      return pytd.NamedType("typing.Callable")
    elif isinstance(v, abstract.FinalAnnotation):
      param = self.value_to_pytd_type(node, v.annotation, seen, view)
      return pytd.GenericType(base_type=pytd.NamedType("typing.Final"),
                              parameters=(param,))
    else:
      raise NotImplementedError(v.__class__.__name__)

  def signature_to_callable(self, sig):
    """Converts a function.Signature object into a callable object.

    Args:
      sig: The signature to convert.

    Returns:
      An abstract.CallableClass representing the signature, or an
      abstract.ParameterizedClass if the signature has a variable number of
      arguments.
    """
    base_cls = self.ctx.convert.function_type
    ret = sig.annotations.get("return", self.ctx.convert.unsolvable)
    if not sig.kwonly_params and (self._detailed or (
        sig.mandatory_param_count() == sig.maximum_param_count())):
      # If self._detailed is false, we throw away the argument types if the
      # function takes a variable number of arguments, which is correct for pyi
      # generation but undesirable for, say, error message printing.
      args = [
          sig.annotations.get(name, self.ctx.convert.unsolvable)
          for name in sig.param_names
      ]
      params = {
          abstract_utils.ARGS: self.ctx.convert.merge_values(args),
          abstract_utils.RET: ret
      }
      params.update(enumerate(args))
      return abstract.CallableClass(base_cls, params, self.ctx)
    else:
      # The only way to indicate kwonly arguments or a variable number of
      # arguments in a Callable is to not specify argument types at all.
      params = {
          abstract_utils.ARGS: self.ctx.convert.unsolvable,
          abstract_utils.RET: ret
      }
      return abstract.ParameterizedClass(base_cls, params, self.ctx)

  def value_to_pytd_def(self, node, v, name):
    """Get a PyTD definition for this object.

    Args:
      node: The node.
      v: The object.
      name: The object name.

    Returns:
      A PyTD definition.
    """
    if isinstance(v, abstract.Module):
      return pytd.Alias(name, pytd.Module(name, module_name=v.full_name))
    elif isinstance(v, abstract.BoundFunction):
      d = self.value_to_pytd_def(node, v.underlying, name)
      assert isinstance(d, pytd.Function)
      sigs = tuple(sig.Replace(params=sig.params[1:]) for sig in d.signatures)
      return d.Replace(signatures=sigs)
    elif isinstance(v, attr_overlay.Attrs):
      ret = pytd.NamedType("typing.Callable")
      md = metadata.to_pytd(v.to_metadata())
      return pytd.Annotated(ret, ("'pytype_metadata'", md))
    elif (isinstance(v, abstract.PyTDFunction) and
          not isinstance(v, typing_overlay.TypeVar)):
      return pytd.Function(
          name=name,
          signatures=tuple(sig.pytd_sig for sig in v.signatures),
          kind=v.kind,
          flags=pytd.MethodFlag.abstract_flag(v.is_abstract))
    elif isinstance(v, abstract.InterpreterFunction):
      return self._function_to_def(node, v, name)
    elif isinstance(v, abstract.SimpleFunction):
      return self._simple_func_to_def(node, v, name)
    elif isinstance(v, (abstract.ParameterizedClass, abstract.Union)):
      return pytd.Alias(name, v.get_instance_type(node))
    elif isinstance(v, abstract.PyTDClass) and v.module:
      # This happens if a module does e.g. "from x import y as z", i.e., copies
      # something from another module to the local namespace. We *could*
      # reproduce the entire class, but we choose a more dense representation.
      return v.to_type(node)
    elif isinstance(v, typed_dict.TypedDictClass):
      return self._typed_dict_to_def(node, v, name)
    elif isinstance(v, abstract.PyTDClass):  # a namedtuple instance
      assert name != v.name
      return pytd.Alias(name, pytd.NamedType(v.name))
    elif isinstance(v, abstract.InterpreterClass):
      if (v.official_name is None or name == v.official_name) and not v.module:
        return self._class_to_def(node, v, name)
      else:
        # Represent a class alias as X: Type[Y] rather than X = Y so the pytd
        # printer can distinguish it from a module alias.
        type_name = v.full_name if v.module else v.official_name
        return pytd.Constant(
            name, pytd.GenericType(pytd.NamedType("builtins.type"),
                                   (pytd.NamedType(type_name),)))
    elif isinstance(v, abstract.TypeParameter):
      return self._typeparam_to_def(node, v, name)
    elif isinstance(v, abstract.Unsolvable):
      return pytd.Constant(name, v.to_type(node))
    else:
      raise NotImplementedError(v.__class__.__name__)

  def _ordered_attrs_to_instance_types(self, node, attr_metadata, annots):
    """Get instance types for ordered attrs in the metadata."""
    attrs = attr_metadata.get("attr_order", [])
    if not annots or not attrs:
      return

    # Use the ordering from attr_order, but use the types in the annotations
    # dict, to handle InitVars correctly (an InitVar without a default will be
    # in attr_order, but not in annotations, and an InitVar with a default will
    # have its type in attr_order set to the inner type).
    annotations = dict(annots.get_annotations(node))
    for a in attrs:
      if a.name in annotations:
        typ = annotations[a.name]
      elif a.kind == class_mixin.AttributeKinds.INITVAR:
        # Do not output initvars without defaults
        typ = None
      else:
        typ = a.typ
      typ = typ and typ.get_instance_type(node)
      yield a.name, typ

  def annotations_to_instance_types(self, node, annots):
    """Get instance types for annotations not present in the members map."""
    if annots:
      for name, local in annots.annotated_locals.items():
        typ = local.get_type(node, name)
        if typ:
          t = typ.get_instance_type(node)
          if local.final:
            t = pytd.GenericType(pytd.NamedType("typing.Final"), (t,))
          yield name, t

  def _function_call_to_return_type(self, node, v, seen_return, num_returns):
    """Get a function call's pytd return type."""
    if v.signature.has_return_annotation:
      if v.is_coroutine():
        ret = abstract.Coroutine.make(self.ctx, v, node).to_type(node)
      else:
        ret = v.signature.annotations["return"].get_instance_type(node)
    else:
      ret = seen_return.data.to_type(node)
      if isinstance(ret, pytd.NothingType) and num_returns == 1:
        if isinstance(seen_return.data, abstract.Empty):
          ret = pytd.AnythingType()
        else:
          assert isinstance(seen_return.data, typing_overlay.NoReturn)
    return ret

  def _function_call_combination_to_signature(
      self, func, call_combination, num_combinations):
    node_after, combination, return_value = call_combination
    params = []
    for i, (name, kind, optional) in enumerate(func.get_parameters()):
      if i < func.nonstararg_count and name in func.signature.annotations:
        t = func.signature.annotations[name].get_instance_type(node_after)
      else:
        t = combination[name].data.to_type(node_after)
      # Python uses ".0" etc. for the names of parameters that are tuples,
      # like e.g. in: "def f((x,  y), z)".
      params.append(
          pytd.Parameter(name.replace(".", "_"), t, kind, optional, None))
    ret = self._function_call_to_return_type(
        node_after, func, return_value, num_combinations)
    if func.has_varargs():
      if func.signature.varargs_name in func.signature.annotations:
        annot = func.signature.annotations[func.signature.varargs_name]
        typ = annot.get_instance_type(node_after)
      else:
        typ = pytd.NamedType("builtins.tuple")
      starargs = pytd.Parameter(func.signature.varargs_name, typ,
                                pytd.ParameterKind.REGULAR, True, None)
    else:
      starargs = None
    if func.has_kwargs():
      if func.signature.kwargs_name in func.signature.annotations:
        annot = func.signature.annotations[func.signature.kwargs_name]
        typ = annot.get_instance_type(node_after)
      else:
        typ = pytd.NamedType("builtins.dict")
      starstarargs = pytd.Parameter(func.signature.kwargs_name, typ,
                                    pytd.ParameterKind.REGULAR, True, None)
    else:
      starstarargs = None
    return pytd.Signature(
        params=tuple(params),
        starargs=starargs,
        starstarargs=starstarargs,
        return_type=ret,
        exceptions=(),  # TODO(b/159052087): record exceptions
        template=())

  def _function_to_def(self, node, v, function_name):
    """Convert an InterpreterFunction to a PyTD definition."""
    signatures = []
    for func in v.signature_functions():
      combinations = func.get_call_combinations(node)
      num_combinations = len(combinations)
      signatures.extend(
          self._function_call_combination_to_signature(
              func, combination, num_combinations)
          for combination in combinations)
    return pytd.Function(
        name=function_name,
        signatures=tuple(signatures),
        kind=pytd.MethodKind.METHOD,
        flags=pytd.MethodFlag.abstract_flag(v.is_abstract))

  def _simple_func_to_def(self, node, v, name):
    """Convert a SimpleFunction to a PyTD definition."""
    sig = v.signature

    def get_parameter(p, kind):
      return pytd.Parameter(p, sig.annotations[p].get_instance_type(node), kind,
                            p in sig.defaults, None)

    posonly = [
        get_parameter(p, pytd.ParameterKind.POSONLY) for p in sig.posonly_params
    ]
    params = [
        get_parameter(p, pytd.ParameterKind.REGULAR)
        for p in sig.param_names[sig.posonly_count:]
    ]
    kwonly = [
        get_parameter(p, pytd.ParameterKind.KWONLY) for p in sig.kwonly_params
    ]
    if sig.varargs_name:
      star = pytd.Parameter(
          sig.varargs_name,
          sig.annotations[sig.varargs_name].get_instance_type(node),
          pytd.ParameterKind.REGULAR, False, None)
    else:
      star = None
    if sig.kwargs_name:
      starstar = pytd.Parameter(
          sig.kwargs_name,
          sig.annotations[sig.kwargs_name].get_instance_type(node),
          pytd.ParameterKind.REGULAR, False, None)
    else:
      starstar = None
    if sig.has_return_annotation:
      ret_type = sig.annotations["return"].get_instance_type(node)
    else:
      ret_type = pytd.NamedType("builtins.NoneType")
    pytd_sig = pytd.Signature(
        params=tuple(posonly+params+kwonly),
        starargs=star,
        starstarargs=starstar,
        return_type=ret_type,
        exceptions=(),
        template=())
    return pytd.Function(name, (pytd_sig,), pytd.MethodKind.METHOD)

  def _function_to_return_types(self, node, fvar):
    """Convert a function variable to a list of PyTD return types."""
    options = fvar.FilteredData(self.ctx.exitpoint, strict=False)
    if not all(isinstance(o, abstract.Function) for o in options):
      return [pytd.AnythingType()]
    types = []
    for val in options:
      if isinstance(val, abstract.InterpreterFunction):
        combinations = val.get_call_combinations(node)
        for node_after, _, return_value in combinations:
          types.append(self._function_call_to_return_type(
              node_after, val, return_value, len(combinations)))
      elif isinstance(val, abstract.PyTDFunction):
        types.extend(sig.pytd_sig.return_type for sig in val.signatures)
      else:
        types.append(pytd.AnythingType())
    safe_types = []  # types without type parameters
    for t in types:
      params = pytd_utils.GetTypeParameters(t)
      t = t.Visit(visitors.ReplaceTypeParameters(
          {p: p.upper_value for p in params}))
      safe_types.append(t)
    return safe_types

  def _is_instance(self, value, cls_name):
    return (isinstance(value, abstract.Instance) and
            value.cls.full_name == cls_name)

  def _class_to_def(self, node, v, class_name):
    """Convert an InterpreterClass to a PyTD definition."""
    self._scopes.append(class_name)
    methods = {}
    constants = collections.defaultdict(pytd_utils.TypeBuilder)

    annots = abstract_utils.get_annotations_dict(v.members)
    annotated_names = set()

    def add_constants(iterator):
      for name, t in iterator:
        if t is None:
          # Remove the entry from constants
          annotated_names.add(name)
        elif name not in annotated_names:
          constants[name].add_type(t)
          annotated_names.add(name)

    add_constants(self._ordered_attrs_to_instance_types(
        node, v.metadata, annots))
    add_constants(self.annotations_to_instance_types(node, annots))

    def add_final(defn, value):
      if value.final:
        return defn.Replace(flags=defn.flags | pytd.MethodFlag.FINAL)
      else:
        return defn

    def get_decorated_method(name, value, func_slot):
      fvar = getattr(value, func_slot)
      func = abstract_utils.get_atomic_value(fvar, abstract.Function)
      defn = self.value_to_pytd_def(node, func, name)
      defn = defn.Visit(visitors.DropMutableParameters())
      defn = add_final(defn, value)
      return defn

    def add_decorated_method(name, value, kind):
      try:
        defn = get_decorated_method(name, value, "func")
      except (AttributeError, abstract_utils.ConversionError):
        constants[name].add_type(pytd.AnythingType())
        return
      defn = defn.Replace(kind=kind)
      methods[name] = defn

    # If decorators are output as aliases to NamedTypes, they will be converted
    # to Functions and fail a verification step if those functions have type
    # parameters. Since we just want the function name, and since we have a
    # fully resolved name at this stage, we just output a minimal pytd.Function
    sig = pytd.Signature((), None, None, pytd.AnythingType(), (), ())
    decorators = [
        pytd.Alias(x, pytd.Function(x, (sig,), pytd.MethodKind.METHOD,
                                    pytd.MethodFlag.NONE))
        for x in v.decorators
    ]
    if v.final:
      fn = pytd.Function("typing.final", (sig,), pytd.MethodKind.METHOD,
                         pytd.MethodFlag.NONE)
      decorators.append(pytd.Alias("final", fn))

    # Collect nested classes
    classes = [self._class_to_def(node, x, x.name)
               for x in v.get_inner_classes()]
    inner_class_names = {x.name for x in classes}

    # class-level attributes
    for name, member in v.members.items():
      if (name in abstract_utils.CLASS_LEVEL_IGNORE or
          name in annotated_names or
          (v.is_enum and name in ("__new__", "__eq__")) or
          name in inner_class_names):
        continue
      for value in member.FilteredData(self.ctx.exitpoint, strict=False):
        if isinstance(value, special_builtins.PropertyInstance):
          # For simplicity, output properties as constants, since our parser
          # turns them into constants anyway.
          if value.fget:
            for typ in self._function_to_return_types(node, value.fget):
              constants[name].add_type(pytd.Annotated(typ, ("'property'",)))
          else:
            constants[name].add_type(
                pytd.Annotated(pytd.AnythingType(), ("'property'",)))
        elif isinstance(value, special_builtins.StaticMethodInstance):
          add_decorated_method(name, value, pytd.MethodKind.STATICMETHOD)
        elif isinstance(value, special_builtins.ClassMethodInstance):
          add_decorated_method(name, value, pytd.MethodKind.CLASSMETHOD)
        elif isinstance(value, abstract.Function):
          # value_to_pytd_def returns different pytd node types depending on the
          # input type, which pytype struggles to reason about.
          method = cast(pytd.Function,
                        self.value_to_pytd_def(node, value, name))
          def fix(sig):
            if not sig.params:
              return sig
            # Check whether the signature's 'self' type is the current class.
            self_type = sig.params[0].type
            maybe_params = pytd_utils.UnpackGeneric(self_type, "builtins.type")
            if maybe_params:
              self_type_name = maybe_params[0].name
            else:
              self_type_name = self_type.name
            if not self_type_name:
              return sig
            full_name = v.official_name or v.name
            if not self_type_name.startswith(full_name):
              return None
            # Remove any outer class prefixes from the type name.
            if "." in full_name:
              new_self_type = self_type.Replace(name=v.name)
              new_first_param = sig.params[0].Replace(type=new_self_type)
              return sig.Replace(params=(new_first_param,) + sig.params[1:])
            else:
              return sig
          signatures = tuple(filter(None, (fix(s) for s in method.signatures)))
          if signatures and signatures != method.signatures:
            # Filter out calls made from subclasses unless they are the only
            # ones recorded; when inferring types for ParentClass.__init__, we
            # do not want `self: Union[ParentClass, Subclass]`.
            method = method.Replace(signatures=signatures)
          method = add_final(method, value)
          # TODO(rechen): Removing mutations altogether won't work for generic
          # classes. To support those, we'll need to change the mutated type's
          # base to the current class, rename aliased type parameters, and
          # replace any parameter not in the class or function template with
          # its upper value.
          methods[name] = method.Visit(visitors.DropMutableParameters())
        elif v.is_enum and self.ctx.options.use_enum_overlay:
          if (any(
              isinstance(enum_member, abstract.Instance) and
              enum_member.cls == v for enum_member in member.data)):
            # i.e. if this is an enum that has any enum members, and the current
            # member is an enum member.
            # In this case, we would normally output:
            # class M(enum.Enum):
            #   A: M
            # However, this loses the type of A.value. Instead, annotate members
            # with the type of their value. (This is what typeshed does.)
            # class M(enum.Enum):
            #   A: int
            enum_member = abstract_utils.get_atomic_value(member)
            node, attr_var = self.ctx.attribute_handler.get_attribute(
                node, enum_member, "value")
            attr = abstract_utils.get_atomic_value(attr_var)
            constants[name].add_type(attr.to_type(node))
          else:
            # i.e. this is an enum, and the current member is NOT an enum
            # member. Which means it's a ClassVar.
            cls_member = abstract_utils.get_atomic_value(member)
            constants[name].add_type(
                pytd.GenericType(
                    base_type=pytd.NamedType("typing.ClassVar"),
                    parameters=((cls_member.to_type(node),))))
        else:
          cls = self.ctx.convert.merge_classes([value])
          node, attr = self.ctx.attribute_handler.get_attribute(
              node, cls, "__get__")
          if attr:
            # This attribute is a descriptor. Its type is the return value of
            # its __get__ method.
            for typ in self._function_to_return_types(node, attr):
              constants[name].add_type(typ)
          else:
            constants[name].add_type(value.to_type(node))

    # Instance-level attributes: all attributes from 'canonical' instances (that
    # is, ones created by analyze.py:analyze_class()) are added. Attributes from
    # non-canonical instances are added if their canonical values do not contain
    # type parameters.
    ignore = set(annotated_names)
    # enums should not print "name" and "value" for instances.
    if v.is_enum:
      ignore.update(("name", "_name_", "value", "_value_"))
    canonical_attributes = set()

    def add_attributes_from(instance):
      for name, member in instance.members.items():
        if name in abstract_utils.CLASS_LEVEL_IGNORE or name in ignore:
          continue
        for value in member.FilteredData(self.ctx.exitpoint, strict=False):
          typ = value.to_type(node)
          if pytd_utils.GetTypeParameters(typ):
            # This attribute's type comes from an annotation that contains a
            # type parameter; we do not want to merge in substituted values of
            # the type parameter.
            canonical_attributes.add(name)
          if v.is_enum:
            # If the containing class (v) is an enum, then output the instance
            # attributes as properties.
            # https://typing.readthedocs.io/en/latest/stubs.html#enums
            typ = pytd.Annotated(typ, ("'property'",))
          constants[name].add_type(typ)

    for instance in v.canonical_instances:
      add_attributes_from(instance)
    ignore |= canonical_attributes
    for instance in v.instances - v.canonical_instances:
      add_attributes_from(instance)

    for name in list(methods):
      if name in constants:
        # If something is both a constant and a method, it means that the class
        # is, at some point, overwriting its own methods with an attribute.
        del methods[name]
        constants[name].add_type(pytd.AnythingType())

    if isinstance(v, named_tuple.NamedTupleClass):
      # Filter out generated members from namedtuples
      cls_bases = v.props.bases
      fieldnames = [x.name for x in v.props.fields]
      methods = {k: m for k, m in methods.items()
                 if k not in v.generated_members}
      constants = {k: c for k, c in constants.items()
                   if k not in v.generated_members}
      for k, c in constants.items():
        # If we have added class members to a namedtuple, do not emit them as
        # regular fields.
        if k not in fieldnames:
          c.wrap("typing.ClassVar")
      slots = None
    else:
      cls_bases = v.bases()
      slots = v.slots

    metaclass = v.metaclass(node)
    if metaclass is not None:
      metaclass = metaclass.get_instance_type(node)

    # Some of the class's bases may not be in global scope, so they won't show
    # up in the output. In that case, fold the base class's type information
    # into this class's pytd.
    bases = []
    missing_bases = []

    for basevar in cls_bases:
      if len(basevar.bindings) == 1:
        b, = basevar.data
        if b.official_name is None and isinstance(b, abstract.InterpreterClass):
          missing_bases.append(b)
        else:
          bases.append(b.get_instance_type(node))
      else:
        bases.append(pytd_utils.JoinTypes(b.get_instance_type(node)
                                          for b in basevar.data))

    # If a namedtuple was constructed via one of the functional forms, it will
    # not have a base class. Since we uniformly output all namedtuple classes as
    # subclasses of typing.NamedTuple we need to add it in here.
    if isinstance(v, named_tuple.NamedTupleClass):
      if not any(x.name == "typing.NamedTuple" for x in bases):
        bases.append(pytd.NamedType("typing.NamedTuple"))

    has_namedtuple_parent = False
    parent_field_names = set()
    for x in missing_bases:
      if isinstance(x, named_tuple.NamedTupleClass):
        has_namedtuple_parent = True
        parent_field_names.update(field.name for field in x.props.fields)
    if has_namedtuple_parent:
      # If inheriting from an anonymous namedtuple, mark all derived class
      # constants as ClassVars, otherwise MergeBaseClasses will convert them
      # into namedtuple fields.
      for k, c in constants.items():
        if k not in parent_field_names:
          c.wrap("typing.ClassVar")

    final_constants = []
    if isinstance(v, named_tuple.NamedTupleClass):
      # The most precise way to get defaults is to check v.__new__.__defaults__,
      # since it's possible for the user to manually set __defaults__. If
      # retrieving this attribute fails, we fall back to the defaults set when
      # the class was built.
      try:
        new = abstract_utils.get_atomic_value(
            v.members["__new__"], abstract.SignedFunction)
      except abstract_utils.ConversionError:
        fields_with_defaults = {f.name for f in v.props.fields if f.default}
      else:
        fields_with_defaults = set(new.signature.defaults)
    else:
      fields_with_defaults = set()
    for name, builder in constants.items():
      if not builder:
        continue
      value = pytd.AnythingType() if name in fields_with_defaults else None
      final_constants.append(pytd.Constant(name, builder.build(), value))

    cls = pytd.Class(name=class_name,
                     metaclass=metaclass,
                     bases=tuple(bases),
                     methods=tuple(methods.values()),
                     constants=tuple(final_constants),
                     classes=tuple(classes),
                     decorators=tuple(decorators),
                     slots=slots,
                     template=())
    for base in missing_bases:
      base_cls = self.value_to_pytd_def(node, base, base.name)
      cls = pytd_utils.MergeBaseClass(cls, base_cls)
    self._scopes.pop()
    return cls

  def _typeparam_to_def(self, node, v, name):
    constraints = tuple(c.get_instance_type(node) for c in v.constraints)
    bound = v.bound and v.bound.get_instance_type(node)
    return pytd.TypeParameter(name, constraints=constraints, bound=bound)

  def _typed_dict_to_def(self, node, v, name):
    constants = []
    for k, var in v.props.fields.items():
      typ = pytd_utils.JoinTypes(
          self.value_instance_to_pytd_type(node, p, None, set(), {})
          for p in var.data)
      constants.append(pytd.Constant(k, typ))
    bases = (pytd.NamedType("typing.TypedDict"),)
    return pytd.Class(name=name,
                      metaclass=None,
                      bases=bases,
                      methods=(),
                      constants=tuple(constants),
                      classes=(),
                      decorators=(),
                      slots=None,
                      template=())



from abc import abstractmethod
from typing import Any

from ._chart import ChartBase

class _LineChartBase(ChartBase):
    grouping: Any
    varyColors: Any
    ser: Any
    dLbls: Any
    dataLabels: Any
    dropLines: Any
    __elements__: Any
    def __init__(
        self,
        grouping: str = ...,
        varyColors: Any | None = ...,
        ser=...,
        dLbls: Any | None = ...,
        dropLines: Any | None = ...,
        **kw,
    ) -> None: ...
    @property
    @abstractmethod
    def tagname(self) -> str: ...

class LineChart(_LineChartBase):
    tagname: str
    grouping: Any
    varyColors: Any
    ser: Any
    dLbls: Any
    dropLines: Any
    hiLowLines: Any
    upDownBars: Any
    marker: Any
    smooth: Any
    extLst: Any
    x_axis: Any
    y_axis: Any
    __elements__: Any
    def __init__(
        self,
        hiLowLines: Any | None = ...,
        upDownBars: Any | None = ...,
        marker: Any | None = ...,
        smooth: Any | None = ...,
        extLst: Any | None = ...,
        **kw,
    ) -> None: ...

class LineChart3D(_LineChartBase):
    tagname: str
    grouping: Any
    varyColors: Any
    ser: Any
    dLbls: Any
    dropLines: Any
    gapDepth: Any
    hiLowLines: Any
    upDownBars: Any
    marker: Any
    smooth: Any
    extLst: Any
    x_axis: Any
    y_axis: Any
    z_axis: Any
    __elements__: Any
    def __init__(
        self,
        gapDepth: Any | None = ...,
        hiLowLines: Any | None = ...,
        upDownBars: Any | None = ...,
        marker: Any | None = ...,
        smooth: Any | None = ...,
        **kw,
    ) -> None: ...



import sys
from typing import NoReturn, Sequence

ACCESS_DEFAULT: int
ACCESS_READ: int
ACCESS_WRITE: int
ACCESS_COPY: int

ALLOCATIONGRANULARITY: int

if sys.platform == "linux":
    MAP_DENYWRITE: int
    MAP_EXECUTABLE: int

if sys.platform != "win32":
    MAP_ANON: int
    MAP_ANONYMOUS: int
    MAP_PRIVATE: int
    MAP_SHARED: int
    PROT_EXEC: int
    PROT_READ: int
    PROT_WRITE: int

    PAGESIZE: int

class mmap(Sequence[bytes]):
    if sys.platform == "win32":
        def __init__(self, fileno: int, length: int, tagname: str | None = ..., access: int = ..., offset: int = ...) -> None: ...
    else:
        def __init__(
            self, fileno: int, length: int, flags: int = ..., prot: int = ..., access: int = ..., offset: int = ...
        ) -> None: ...

    def close(self) -> None: ...
    def flush(self, offset: int = ..., size: int = ...) -> int: ...
    def move(self, dest: int, src: int, count: int) -> None: ...
    def read_byte(self) -> bytes: ...
    def readline(self) -> bytes: ...
    def resize(self, newsize: int) -> None: ...
    def seek(self, pos: int, whence: int = ...) -> None: ...
    def size(self) -> int: ...
    def tell(self) -> int: ...
    def write_byte(self, byte: bytes) -> None: ...
    def __len__(self) -> int: ...
    def find(self, string: bytes, start: int = ..., end: int = ...) -> int: ...
    def rfind(self, string: bytes, start: int = ..., stop: int = ...) -> int: ...
    def read(self, num: int) -> bytes: ...
    def write(self, string: bytes) -> None: ...
    def __getitem__(self, index: int | slice) -> bytes: ...
    def __getslice__(self, i: int | None, j: int | None) -> bytes: ...
    def __delitem__(self, index: int | slice) -> NoReturn: ...
    def __setitem__(self, index: int | slice, object: bytes) -> None: ...



from abc import ABCMeta, abstractmethod
from typing import Any

from cryptography.hazmat.primitives.asymmetric.dh import (
    DHParameterNumbers,
    DHParameters,
    DHPrivateKey,
    DHPrivateNumbers,
    DHPublicKey,
    DHPublicNumbers,
)
from cryptography.hazmat.primitives.asymmetric.dsa import (
    DSAParameterNumbers,
    DSAParameters,
    DSAPrivateKey,
    DSAPrivateNumbers,
    DSAPublicKey,
    DSAPublicNumbers,
)
from cryptography.hazmat.primitives.asymmetric.ec import (
    EllipticCurve,
    EllipticCurvePrivateKey,
    EllipticCurvePrivateNumbers,
    EllipticCurvePublicKey,
    EllipticCurvePublicNumbers,
    EllipticCurveSignatureAlgorithm,
)
from cryptography.hazmat.primitives.asymmetric.padding import AsymmetricPadding
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey, RSAPrivateNumbers, RSAPublicKey, RSAPublicNumbers
from cryptography.hazmat.primitives.ciphers import BlockCipherAlgorithm, CipherAlgorithm, CipherContext
from cryptography.hazmat.primitives.ciphers.modes import Mode
from cryptography.hazmat.primitives.hashes import HashAlgorithm, HashContext
from cryptography.x509 import (
    Certificate,
    CertificateBuilder,
    CertificateRevocationList,
    CertificateRevocationListBuilder,
    CertificateSigningRequest,
    CertificateSigningRequestBuilder,
    Name,
    RevokedCertificate,
    RevokedCertificateBuilder,
)

class CipherBackend(metaclass=ABCMeta):
    @abstractmethod
    def cipher_supported(self, cipher: CipherAlgorithm, mode: Mode) -> bool: ...
    @abstractmethod
    def create_symmetric_encryption_ctx(self, cipher: CipherAlgorithm, mode: Mode) -> CipherContext: ...
    @abstractmethod
    def create_symmetric_decryption_ctx(self, cipher: CipherAlgorithm, mode: Mode) -> CipherContext: ...

class CMACBackend(metaclass=ABCMeta):
    @abstractmethod
    def cmac_algorithm_supported(self, algorithm: BlockCipherAlgorithm) -> bool: ...
    @abstractmethod
    def create_cmac_ctx(self, algorithm: BlockCipherAlgorithm) -> Any: ...

class DERSerializationBackend(metaclass=ABCMeta):
    @abstractmethod
    def load_der_parameters(self, data: bytes) -> Any: ...
    @abstractmethod
    def load_der_private_key(self, data: bytes, password: bytes | None) -> Any: ...
    @abstractmethod
    def load_der_public_key(self, data: bytes) -> Any: ...

class DHBackend(metaclass=ABCMeta):
    @abstractmethod
    def dh_parameters_supported(self, p: int, g: int, q: int | None) -> bool: ...
    @abstractmethod
    def dh_x942_serialization_supported(self) -> bool: ...
    @abstractmethod
    def generate_dh_parameters(self, generator: int, key_size: int) -> DHParameters: ...
    @abstractmethod
    def generate_dh_private_key(self, parameters: DHParameters) -> DHPrivateKey: ...
    @abstractmethod
    def generate_dh_private_key_and_parameters(self, generator: int, key_size: int) -> DHPrivateKey: ...
    @abstractmethod
    def load_dh_parameter_numbers(self, numbers: DHParameterNumbers) -> DHParameters: ...
    @abstractmethod
    def load_dh_private_numbers(self, numbers: DHPrivateNumbers) -> DHPrivateKey: ...
    @abstractmethod
    def load_dh_public_numbers(self, numbers: DHPublicNumbers) -> DHPublicKey: ...

class DSABackend(metaclass=ABCMeta):
    @abstractmethod
    def dsa_hash_supported(self, algorithm: HashAlgorithm) -> bool: ...
    @abstractmethod
    def dsa_parameters_supported(self, p: int, q: int, g: int) -> bool: ...
    @abstractmethod
    def generate_dsa_parameters(self, key_size: int) -> DSAParameters: ...
    @abstractmethod
    def generate_dsa_private_key(self, parameters: DSAParameters) -> DSAPrivateKey: ...
    @abstractmethod
    def generate_dsa_private_key_and_parameters(self, key_size: int) -> DSAPrivateKey: ...
    @abstractmethod
    def load_dsa_parameter_numbers(self, numbers: DSAParameterNumbers) -> DSAParameters: ...
    @abstractmethod
    def load_dsa_private_numbers(self, numbers: DSAPrivateNumbers) -> DSAPrivateKey: ...
    @abstractmethod
    def load_dsa_public_numbers(self, numbers: DSAPublicNumbers) -> DSAPublicKey: ...

class EllipticCurveBackend(metaclass=ABCMeta):
    @abstractmethod
    def derive_elliptic_curve_private_key(self, private_value: int, curve: EllipticCurve) -> EllipticCurvePrivateKey: ...
    @abstractmethod
    def elliptic_curve_signature_algorithm_supported(
        self, signature_algorithm: EllipticCurveSignatureAlgorithm, curve: EllipticCurve
    ) -> bool: ...
    @abstractmethod
    def elliptic_curve_supported(self, curve: EllipticCurve) -> bool: ...
    @abstractmethod
    def generate_elliptic_curve_private_key(self, curve: EllipticCurve) -> EllipticCurvePrivateKey: ...
    @abstractmethod
    def load_elliptic_curve_private_numbers(self, numbers: EllipticCurvePrivateNumbers) -> EllipticCurvePrivateKey: ...
    @abstractmethod
    def load_elliptic_curve_public_numbers(self, numbers: EllipticCurvePublicNumbers) -> EllipticCurvePublicKey: ...

class HMACBackend(metaclass=ABCMeta):
    @abstractmethod
    def create_hmac_ctx(self, key: bytes, algorithm: HashAlgorithm) -> HashContext: ...
    @abstractmethod
    def cmac_algorithm_supported(self, algorithm: HashAlgorithm) -> bool: ...

class HashBackend(metaclass=ABCMeta):
    @abstractmethod
    def create_hash_ctx(self, algorithm: HashAlgorithm) -> HashContext: ...
    @abstractmethod
    def hash_supported(self, algorithm: HashAlgorithm) -> bool: ...

class PBKDF2HMACBackend(metaclass=ABCMeta):
    @abstractmethod
    def derive_pbkdf2_hmac(
        self, algorithm: HashAlgorithm, length: int, salt: bytes, iterations: int, key_material: bytes
    ) -> bytes: ...
    @abstractmethod
    def pbkdf2_hmac_supported(self, algorithm: HashAlgorithm) -> bool: ...

class PEMSerializationBackend(metaclass=ABCMeta):
    @abstractmethod
    def load_pem_parameters(self, data: bytes) -> Any: ...
    @abstractmethod
    def load_pem_private_key(self, data: bytes, password: bytes | None) -> Any: ...
    @abstractmethod
    def load_pem_public_key(self, data: bytes) -> Any: ...

class RSABackend(metaclass=ABCMeta):
    @abstractmethod
    def generate_rsa_parameters_supported(self, public_exponent: int, key_size: int) -> bool: ...
    @abstractmethod
    def generate_rsa_private_key(self, public_exponent: int, key_size: int) -> RSAPrivateKey: ...
    @abstractmethod
    def load_rsa_public_numbers(self, numbers: RSAPublicNumbers) -> RSAPublicKey: ...
    @abstractmethod
    def load_rsa_private_numbers(self, numbers: RSAPrivateNumbers) -> RSAPrivateKey: ...
    @abstractmethod
    def rsa_padding_supported(self, padding: AsymmetricPadding) -> bool: ...

class ScryptBackend(metaclass=ABCMeta):
    @abstractmethod
    def derive_scrypt(self, key_material: bytes, salt: bytes, length: int, n: int, r: int, p: int) -> bytes: ...

class X509Backend(metaclass=ABCMeta):
    @abstractmethod
    def create_x509_certificate(
        self,
        builder: CertificateBuilder,
        private_key: DSAPrivateKey | EllipticCurvePrivateKey | RSAPrivateKey,
        algorithm: HashAlgorithm,
    ) -> Certificate: ...
    @abstractmethod
    def create_x509_crl(
        self,
        builder: CertificateRevocationListBuilder,
        private_key: DSAPrivateKey | EllipticCurvePrivateKey | RSAPrivateKey,
        algorithm: HashAlgorithm,
    ) -> CertificateRevocationList: ...
    @abstractmethod
    def create_x509_csr(
        self,
        builder: CertificateSigningRequestBuilder,
        private_key: DSAPrivateKey | EllipticCurvePrivateKey | RSAPrivateKey,
        algorithm: HashAlgorithm,
    ) -> CertificateSigningRequest: ...
    @abstractmethod
    def create_x509_revoked_certificate(self, builder: RevokedCertificateBuilder) -> RevokedCertificate: ...
    @abstractmethod
    def load_der_x509_certificate(self, data: bytes) -> Certificate: ...
    @abstractmethod
    def load_der_x509_csr(self, data: bytes) -> CertificateSigningRequest: ...
    @abstractmethod
    def load_pem_x509_certificate(self, data: bytes) -> Certificate: ...
    @abstractmethod
    def load_pem_x509_csr(self, data: bytes) -> CertificateSigningRequest: ...
    @abstractmethod
    def x509_name_bytes(self, name: Name) -> bytes: ...



"""Code for checking and inferring types."""

import dataclasses
import logging

from typing import Optional

from pytype import context
from pytype import convert_structural
from pytype import debug
from pytype import errors
from pytype import metrics
from pytype.abstract import abstract_utils
from pytype.inspect import graph
from pytype.pytd import pytd
from pytype.pytd import pytd_utils
from pytype.pytd import visitors

log = logging.getLogger(__name__)

# How deep to follow call chains:
INIT_MAXIMUM_DEPTH = 4  # during module loading
MAXIMUM_DEPTH = 3  # during non-quick analysis
QUICK_CHECK_MAXIMUM_DEPTH = 2  # during quick checking
QUICK_INFER_MAXIMUM_DEPTH = 1  # during quick inference


@dataclasses.dataclass
class Analysis:
  ast: Optional[pytd.TypeDeclUnit]
  builtins: Optional[pytd.TypeDeclUnit]
  errorlog: errors.ErrorLog


def check_types(src, filename, options, loader, deep=True,
                init_maximum_depth=INIT_MAXIMUM_DEPTH,
                maximum_depth=None, **kwargs):
  """Verify the Python code."""
  ctx = context.Context(
      options=options,
      generate_unknowns=False,
      loader=loader,
      **kwargs)
  loc, defs = ctx.vm.run_program(src, filename, init_maximum_depth)
  snapshotter = metrics.get_metric("memory", metrics.Snapshot)
  snapshotter.take_snapshot("analyze:check_types:tracer")
  if deep:
    if maximum_depth is None:
      maximum_depth = (
          QUICK_CHECK_MAXIMUM_DEPTH if options.quick else MAXIMUM_DEPTH)
    ctx.vm.analyze(loc, defs, maximum_depth=maximum_depth)
  snapshotter.take_snapshot("analyze:check_types:post")
  _maybe_output_debug(options, ctx.program)
  return Analysis(None, None, ctx.errorlog)


def infer_types(src,
                options,
                loader,
                filename=None,
                deep=True,
                init_maximum_depth=INIT_MAXIMUM_DEPTH,
                show_library_calls=False,
                maximum_depth=None,
                ctx=None,
                **kwargs):
  """Given Python source return its types.

  Args:
    src: A string containing Python source code.
    options: config.Options object
    loader: A load_pytd.Loader instance to load PYI information.
    filename: Filename of the program we're parsing.
    deep: If True, analyze all functions, even the ones not called by the main
      execution flow.
    init_maximum_depth: Depth of analysis during module loading.
    show_library_calls: If True, call traces are kept in the output.
    maximum_depth: Depth of the analysis. Default: unlimited.
    ctx: An instance of context.Context, in case the caller wants to
      instantiate and retain the abstract context used for type inference.
    **kwargs: Additional parameters to pass to context.Context
  Returns:
    A tuple of (ast: TypeDeclUnit, builtins: TypeDeclUnit)
  Raises:
    AssertionError: In case of a bad parameter combination.
  """
  if not ctx:
    ctx = context.Context(
        options=options,
        generate_unknowns=options.protocols,
        store_all_calls=not deep,
        loader=loader,
        **kwargs)
  loc, defs = ctx.vm.run_program(src, filename, init_maximum_depth)
  log.info("===Done running definitions and module-level code===")
  snapshotter = metrics.get_metric("memory", metrics.Snapshot)
  snapshotter.take_snapshot("analyze:infer_types:tracer")
  if deep:
    if maximum_depth is None:
      if not options.quick:
        maximum_depth = MAXIMUM_DEPTH
      elif options.analyze_annotated:
        # Since there's no point in analyzing annotated functions for inference,
        # the presence of this option means that the user wants checking, too.
        maximum_depth = QUICK_CHECK_MAXIMUM_DEPTH
      else:
        maximum_depth = QUICK_INFER_MAXIMUM_DEPTH
    ctx.exitpoint = ctx.vm.analyze(loc, defs, maximum_depth)
  else:
    ctx.exitpoint = loc
  snapshotter.take_snapshot("analyze:infer_types:post")
  ast = ctx.vm.compute_types(defs)
  ast = ctx.loader.resolve_ast(ast)
  if ctx.vm.has_unknown_wildcard_imports or any(
      a in defs for a in abstract_utils.DYNAMIC_ATTRIBUTE_MARKERS):
    if "__getattr__" not in ast:
      ast = pytd_utils.Concat(ast, ctx.loader.get_default_ast())
  # If merged with other if statement, triggers a ValueError: Unresolved class
  # when attempts to load from the protocols file
  if options.protocols:
    protocols_pytd = ctx.loader.import_name("protocols")
  else:
    protocols_pytd = None
  builtins_pytd = ctx.loader.concat_all()
  # Insert type parameters, where appropriate
  ast = ast.Visit(visitors.CreateTypeParametersForSignatures())
  if options.protocols:
    log.info("=========== PyTD to solve =============\n%s",
             pytd_utils.Print(ast))
    ast = convert_structural.convert_pytd(ast, builtins_pytd, protocols_pytd)
  elif not show_library_calls:
    log.info("Solving is turned off. Discarding call traces.")
    # Rename remaining "~unknown" to "?"
    ast = ast.Visit(visitors.RemoveUnknownClasses())
    # Remove "~list" etc.:
    ast = convert_structural.extract_local(ast)
  _maybe_output_debug(options, ctx.program)
  return Analysis(ast, builtins_pytd, ctx.errorlog)


def _maybe_output_debug(options, program):
  """Maybe emit debugging output."""
  if options.output_cfg or options.output_typegraph:
    tg = graph.TypeGraph(program, set(), bool(options.output_cfg))
    svg_file = options.output_cfg or options.output_typegraph
    graph.write_svg_from_dot(svg_file, tg.to_dot())

  if options.output_debug:
    text = debug.program_to_text(program)
    if options.output_debug == "-":
      log.info("=========== Program Dump =============\n%s", text)
    else:
      with options.open_function(options.output_debug, "w") as fi:
        fi.write(text)



from typing import Any

log: Any
SERVICE_NAME: str
ORIGIN: str
IMDS_URL: str

def initialize() -> None: ...
def get_token(): ...
def get_metadata(token: Any | None = ...): ...
def parse_metadata_json(json_str): ...
def do_request(url, headers: Any | None = ..., method: str = ...): ...



"""Tests for io.py."""

import contextlib
import io as builtins_io
import os
import sys
import tempfile
import textwrap
import traceback

from pytype import config
from pytype import io
from pytype.pytd import pytd

import unittest


class IOTest(unittest.TestCase):
  """Test IO functions."""

  def test_read_source_file_utf8(self):
    with self._tmpfile("abc□def\n") as f:
      self.assertEqual(io.read_source_file(f.name), "abc□def\n")

  @contextlib.contextmanager
  def _tmpfile(self, contents):
    tempfile_options = {"mode": "w", "suffix": ".txt", "encoding": "utf-8"}
    with tempfile.NamedTemporaryFile(**tempfile_options) as f:
      f.write(contents)
      f.flush()
      yield f

  def test_wrap_pytype_exceptions(self):
    with self.assertRaises(ValueError):
      with io.wrap_pytype_exceptions(ValueError, "foo.py"):
        io.read_source_file("missing_file")

  def test_wrap_pytype_exception_traceback(self):
    class CustomError(Exception):
      pass

    def called_function():
      raise OSError("error!")

    def calling_function():
      called_function()

    err = None
    trace = None
    try:
      with io.wrap_pytype_exceptions(CustomError, "foo.py"):
        calling_function()
    except CustomError as e:
      err = e
      _, _, tb = sys.exc_info()
      trace = traceback.format_tb(tb)

    self.assertIn("OSError: error!", err.args[0])
    self.assertTrue(any("in calling_function" in x for x in trace))

  def test_check_py(self):
    errorlog = io.check_py("undefined_var")
    error, = errorlog.unique_sorted_errors()
    self.assertEqual(error.name, "name-error")

  def test_check_py_with_options(self):
    options = config.Options.create(disable="name-error")
    errorlog = io.check_py("undefined_var", options)
    self.assertFalse(errorlog.unique_sorted_errors())

  def test_generate_pyi(self):
    errorlog, pyi_string, pytd_ast = io.generate_pyi("x = 42")
    self.assertFalse(errorlog.unique_sorted_errors())
    self.assertEqual(pyi_string, "x: int\n")
    self.assertIsInstance(pytd_ast, pytd.TypeDeclUnit)

  def test_generate_pyi_with_options(self):
    with self._tmpfile("x: int") as pyi:
      pyi_name, _ = os.path.splitext(os.path.basename(pyi.name))
      with self._tmpfile(
          f"{pyi_name} {pyi.name}") as imports_map:
        src = "import {mod}; y = {mod}.x".format(mod=pyi_name)
        options = config.Options.create(imports_map=imports_map.name)
        _, pyi_string, _ = io.generate_pyi(src, options)
    self.assertEqual(pyi_string,
                     f"import {pyi_name}\n\ny: int\n")

  def test_generate_pyi__overload_order(self):
    _, pyi_string, _ = io.generate_pyi(textwrap.dedent("""
      from typing import Any, overload
      @overload
      def f(x: None) -> None: ...
      @overload
      def f(x: Any) -> int: ...
      def f(x):
        return __any_object__
    """.lstrip("\n")))
    self.assertMultiLineEqual(pyi_string, textwrap.dedent("""
      from typing import overload

      @overload
      def f(x: None) -> None: ...
      @overload
      def f(x) -> int: ...
    """.lstrip("\n")))

  def test_check_or_generate_pyi__check(self):
    with self._tmpfile("") as f:
      options = config.Options.create(f.name, check=True)
      _, pyi_string, pytd_ast = io.check_or_generate_pyi(options)
    self.assertIsNone(pyi_string)
    self.assertIsNone(pytd_ast)

  def test_check_or_generate_pyi__generate(self):
    with self._tmpfile("") as f:
      options = config.Options.create(f.name, check=False)
      _, pyi_string, pytd_ast = io.check_or_generate_pyi(options)
    self.assertIsNotNone(pyi_string)
    self.assertIsNotNone(pytd_ast)

  def test_check_or_generate_pyi__open_function(self):
    def mock_open(filename, *args, **kwargs):
      if filename == "my_amazing_file.py":
        return builtins_io.StringIO("x = 0.0")
      else:
        return open(filename, *args, **kwargs)  # pylint: disable=consider-using-with
    options = config.Options.create(
        "my_amazing_file.py", check=False, open_function=mock_open)
    _, pyi_string, _ = io.check_or_generate_pyi(options)
    self.assertEqual(pyi_string, "x: float\n")

  def test_write_pickle(self):
    ast = pytd.TypeDeclUnit(None, (), (), (), (), ())
    options = config.Options.create(output="/dev/null")
    io.write_pickle(ast, options)  # just make sure we don't crash


if __name__ == "__main__":
  unittest.main()



from typing import Any, ClassVar
from typing_extensions import Literal

from .JpegImagePlugin import JpegImageFile

class MpoImageFile(JpegImageFile):
    format: ClassVar[Literal["MPO"]]
    def load_seek(self, pos) -> None: ...
    fp: Any
    offset: Any
    tile: Any
    def seek(self, frame) -> None: ...
    def tell(self): ...
    @staticmethod
    def adopt(jpeg_instance, mpheader: Any | None = ...): ...



from _typeshed import Self
from collections.abc import Callable, Iterable
from typing import Any, TypeVar, overload

from .config import Config
from .context import Context
from .parser import Argument

_TaskT = TypeVar("_TaskT", bound=Task)

NO_DEFAULT: object

class Task:
    body: Callable[..., Any]
    __doc__: str | None
    __name__: str
    __module__: str
    aliases: tuple[str, ...]
    is_default: bool
    positional: Iterable[str]
    optional: Iterable[str]
    iterable: Iterable[str]
    incrementable: Iterable[str]
    auto_shortflags: bool
    help: dict[str, str]
    pre: Iterable[Task]
    post: Iterable[Task]
    times_called: int
    autoprint: bool
    def __init__(
        self,
        body: Callable[..., Any],
        name: str | None = ...,
        aliases: tuple[str, ...] = ...,
        positional: Iterable[str] | None = ...,
        optional: Iterable[str] = ...,
        default: bool = ...,
        auto_shortflags: bool = ...,
        help: dict[str, str] | None = ...,
        pre: Iterable[Task] | None = ...,
        post: Iterable[Task] | None = ...,
        autoprint: bool = ...,
        iterable: Iterable[str] | None = ...,
        incrementable: Iterable[str] | None = ...,
    ) -> None: ...
    @property
    def name(self): ...
    def __eq__(self, other: Task) -> bool: ...  # type: ignore[override]
    def __hash__(self) -> int: ...
    def __call__(self, *args, **kwargs): ...
    @property
    def called(self) -> bool: ...
    def argspec(self, body): ...
    def fill_implicit_positionals(self, positional: Iterable[str] | None) -> Iterable[str]: ...
    def arg_opts(self, name: str, default: Any, taken_names: Iterable[str]) -> dict[str, Any]: ...
    def get_arguments(self) -> list[Argument]: ...

@overload
def task(__func: Callable[..., Any]) -> Task: ...
@overload
def task(
    *args: Task,
    name: str | None = ...,
    aliases: tuple[str, ...] = ...,
    positional: Iterable[str] | None = ...,
    optional: Iterable[str] = ...,
    default: bool = ...,
    auto_shortflags: bool = ...,
    help: dict[str, str] | None = ...,
    pre: list[Task] | None = ...,
    post: list[Task] | None = ...,
    autoprint: bool = ...,
    iterable: Iterable[str] | None = ...,
    incrementable: Iterable[str] | None = ...,
    klass: type[_TaskT] = ...,
) -> Callable[[Callable[..., Any]], _TaskT]: ...

class Call:
    task: Task
    called_as: str | None
    args: tuple[Any, ...]
    kwargs: dict[str, Any]
    def __init__(
        self, task: Task, called_as: str | None = ..., args: tuple[Any, ...] | None = ..., kwargs: dict[str, Any] | None = ...
    ) -> None: ...
    def __getattr__(self, name: str) -> Any: ...
    def __deepcopy__(self: Self, memo: Any) -> Self: ...
    def __eq__(self, other: Call) -> bool: ...  # type: ignore[override]
    def make_context(self, config: Config) -> Context: ...
    def clone_data(self): ...
    # TODO use overload
    def clone(self, into: type[Call] | None = ..., with_: dict[str, Any] | None = ...) -> Call: ...

def call(task: Task, *args: Any, **kwargs: Any) -> Call: ...
